/*
Copyright (c). The Regents of the
University of California (Regents). All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby granted,
provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
for commercial licensing opportunities. Created by Eric C. Anderson and
John Novembre,
Department of Integrative Biology, University of California, Berkeley.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/




/*  From: novembre@socrates.Berkeley.EDU
To: "Eric C. Anderson" <dr_eriq@uclink4.berkeley.edu>
Subject: DP_Blocks.c
Date: 21 May 2002 14:07:43 -0700
Lines: 375

Eric added some options for different types of Priors on the q's (i.e. gQPrior).
	These options take the form of -q immediately followed (no spaces!!) by another
	letter or number that gives the particular option.  The numbers correspond to the
	numbers in the {enumerate} environment in the manuscript where the distributions are
	described:
			
			-qc	or -q1	-->  the "standard" model, each haplotype uniform conditional on the
							 preceding ones
			-qe	or -q2	-->  multivariate Ewens Distribution
			
			-qo	or -q3	-->  "occupancy" distribution that favors most haplotypes having
						 comparable frequencies (kind of silly)
						 
			-qm or -q0	-->	do a "mixture" of them all.  i.e., take -log_2(3) bits to
							specify which prior is being used for each block, and use the
							one for each block that minimizes the cost.  Cool, No?!  
	
	THE DEFAULT WILL BE -qm

*/


#define UN_EXTERN
//#define MAC_THING
#define RELEASE_VERSION

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>


#ifdef __MWERKS__
	#ifdef macintosh
		#include <console.h>
	#endif
#endif


#include "ECA_MemAlloc.h"
#include "MathStatRand.h"

#include "MDL_Blocks.h"
#include "ranlib.h"


int main(int argc,char** argv)
{	
	MDL_struct *B;	
	int i,j;
	
	// ScoreBoundsMatrix is a ragged matrix where each row is a set of bounds that will be scored 
	// The matrix is filled in from a file specified by the user at the command-line
	// and gives the user a way of scoring arbitrary sets of bounds
	int nScoreBoundsSets=0;
	int *nScoreBounds;
	int **ScoreBoundsMatrix;
	char ScoreBoundsFileName[100];

	//int LookingForArgC = 3;
	FILE *LogFile;
	char SeedFileName[100], LogFileName[100], InfileName[100];

	long seed1, seed2;
	time_t StartTime;
	int ref;
	int DoWangMethod = 0, ComputeHeterozyg=0;
	int OutPutData=0;
	int *Rvisited;
	double *Rscores;
	double IncompThresh=0.0;  /* the proportion of incompatible haplotypes required for the 
							WangMethod with slop to call it a break point */

#ifdef __MWERKS__
	#ifdef macintosh
		argc = ccommand(&argv);
	#endif
#endif	
	StartTime = time(NULL);
	
	// set a random seed.  Typically we will overwrite these, though.
	seed1 = (long)StartTime / 834  + 7737;
	seed2 = (seed1 + 438977)  / 264;
	
	// Set some defaults
	gGreedy=0;
	gVerboseFlag=1;
	gQPrior = MIXTURE;
	gUseKMeansClustering = 0;

	if(argc<2)
	  usage();
	else{
	  sprintf(InfileName,"%s",argv[1]);
	  sprintf(LogFileName,"%s.log",argv[1]);
	  sprintf(SeedFileName,"%s.seeds",argv[1]);
	}
	i=2;
	while(i<argc){
	  switch(argv[i][1]){
	  case 's':
	    i++;
	    seed1=(long)atoi(argv[i]);
	    seed2=(long)atoi(argv[i+1]);
	    i+=2;
	    break;
	  case 'g':
	    gGreedy=1;
	    i++;
	    break;
#ifndef RELEASE_VERSION
	  case 'l':	  
	    i++;
	    sprintf(LogFileName,"%s",argv[i]);
	    i++;
	    break;
	  case 'b':
	    i++;
	    sprintf(SeedFileName,"%s",argv[i]);
	    // get the seeds if the file for them exists and they weren't passed in on 
	    // the command line.  Otherwise the seeds are the default
	    if((SeedFile = fopen(SeedFileName,"r")) != NULL) {
	      fscanf(SeedFile,"%d%d", &seed1,&seed2);
	      fclose(SeedFile);
	    } else{
	      fprintf(stderr,"Couldn't open seed file %s.  Seeds set to %d %d.",SeedFileName,seed1,seed2);
	    }
	    i++;
	    break;
	  case 'c':
	    DoWangMethod=1;
	    i++;
	    break;
	  case 'h':
	    ComputeHeterozyg=1;
	    i++;
	    break;
	  case 't':
	    i++;
	    DoWangMethod=1;
	    IncompThresh=(double)atof(argv[i]);
	    if(IncompThresh<=0.0 || IncompThresh >=.25) {
	    	printf("\nSilly Choice, %f, for -t option.  Should be b/t 0 and .25\n");
	    	usage();
	    }
	    i++;
	    break;
	  case 'q':  /* gQPrior Options */
	  	switch(argv[i][2]) {
	  		case 'c':
	  		case '1':
	  			gQPrior = STANDARD;
	  			break;
	  		case 'e':
	  		case '2':
	  			gQPrior = EWENS;
	  			break;
	  		case 'o':
	  		case '3':
	  			gQPrior = OCCUPANCY;
	  			break;
	  		case 'm':
	  		case '0':
	  			gQPrior = MIXTURE;
	  			break;
	  		default:
	  			fprintf(stderr, "\nUsing the -q option inappropriately!\n\n");
	  			usage(); 
	  	}
	    i++;
	    break;
	  case 'e':
	    gQPrior = EWENS;
	    i++;
	    break;
	  case 'v':
	    i++;
	    gVerboseFlag=atoi(argv[i]);
	    i++;
	    break;
	  case 'B':
	    i++;
	    // Read number of sets of bounds
	    nScoreBoundsSets=atoi(argv[i]);
	    i++;
	    // Read filename for boundary sets
	    sprintf(ScoreBoundsFileName,"%s",argv[i]);
	    i++;
	    break;
	  case 'O':
	    OutPutData=1;
	    i++;
	    break;
	  case 'k':
	  	gUseKMeansClustering = 1;
	  	i++;
	  	break;
#endif  // RELEASE_VERSION
	  default:
	    usage();	    
	  }
	}
	
	printf("Program MDBlocks Version 1.0");
	printf("\nReleased 14 MAY 03");
	printf("\nwritten by Eric C. Anderson (dr_eriq@uclink.berkeley.edu)");
	printf("\n       and John Novembre (novembre@socrates.berkeley.edu)");
	printf("\nCopyright (c) by The Regents of the University of California\n");
	printf("Please see user documentation for full software agreement.");
	
	// seed the RNG
	printf("\n\n\nSeeds: %ld %ld\n",seed1,seed2);	
	setall(seed1, seed2);
	
	
	
	
	
       	
	// open the log file and record this momentous occasion:
	LogFile = fopen(LogFileName,"a");
	fprintf(LogFile, "\nProgram %s Launched %s  \twith options:  ",argv[0], ctime(&StartTime));
	for(i=1;i<argc;i++)  {
		fprintf(LogFile,"%s\t",argv[i]);
	}
	fprintf(LogFile,"\n\tStartingSeeds: %ld  %ld", seed1, seed2);
	
	
	// set some global variables relating to how much output is made
	gCloserLook = 0;
	sprintf(gCloserLookFileName, "BlockSeqs.txt");
/*	gCloserLookFile = fopen(gCloserLookFileName, "w");  // open and close in "w" mode to erase all
	fclose(gCloserLookFile);
*/
#ifndef RELEASE_VERSION
	gPrintTestBoundsFile =  fopen("TestedBounds.txt","w");
#endif
	gPrintTestBounds = 0;
	
	
	// then read the data set:
	B = ReadSeqsAsBits(InfileName);
	sprintf(B->InfileName, "%s", InfileName);
	
	
	
	/********* COMPUTE HETEROZYGOSITY ************/
	if(ComputeHeterozyg)  {
		
		PrintHeterozygosities(B);
		return(0);
	}
	

	/************ ERIC IS TESTING SOME K-MEANS STUFF HERE ************/
/*	while(gUseKMeansClustering==1)  {int aa; int bb; int nn;
	
		printf("\nSorry! kmeans not available now! Exiting to system...");
		 exit(1);  */
/*
		printf("\n\nGive MaxNC, a, and b\n->");
		scanf(" %d %d %d",&nn,&aa,&bb);
		while(nn>=1)  { 
			
			cluster_a2b(B, nn, aa, bb);
			nn--;
		}
*/
		
	/* }   */
	
	
	
	if(gVerboseFlag==5){
	  ref = 0;
	  while(ref>=0) {
	    printf("\nGive me a reference! Neg to quit  \n->");
	    scanf("%d",&ref);
	    for(i=0;i<B->NumMarkers;i++) { 
	      double temp_dp; int ssize;
	      int sp,j;
	      temp_dp = pwise_d_prime(B, i, ref, &ssize);
	      sp = (int) (80 * fabs(temp_dp));
	      printf("\n%d->%d",ref,i);
	      for(j=0;j<sp;j++)  {
		printf("*");
	      }
	      //printf("        %d %.5f (%d)",i,temp_dp,ssize);
	    }
	  }
	  
	  scanf("%d",&i);
	}
	
	// read in the base calls if this was for daly, otherwise, leave gDalyBases as NULL
	gDalyBases=NULL;
	if(strcmp(InfileName, "DalyDataAsBits.dat")==0) { 
		int i;
		FILE *din = fopen("DalyTrueBases.dat","r");
		if(din != NULL)  {
			fscanf(din,"%*d");  // eat the first int
			gDalyBases = (char **)ECA_CALLOC((size_t)B->NumMarkers, sizeof(char *));
			for(i=0;i<B->NumMarkers;i++)  {
				gDalyBases[i] = (char *)ECA_CALLOC( (size_t)2,sizeof(char));
				fscanf(din," %c %c",&gDalyBases[i][0],&gDalyBases[i][1]);
			}
		fclose(din);
		} 

	}
	
	
	/******************  FOR THE -c OPTION WE CALL THE WANG FUNCTION AND THEN RETURN OUT OF MAIN() *********/
	if(DoWangMethod)  {
		if(IncompThresh <= 0.0)
			DoWang(B);
		else
			DoWangWithSlop(B,IncompThresh);
		
		/* fill up some memory so that we can still fill out values in PrintEasyGrepLine */
		gDP_BitScore = 0;
		gKLDs = (double *)ECA_CALLOC( (size_t)gDP_NumBoundsInferred+5, sizeof(double) );
		B->ChosenQPrior = (enum qtype_enum **)ECA_CALLOC((size_t)B->NumMarkers,sizeof(enum qtype_enum *));
		for(j=0;j<B->NumMarkers;j++)  {
			B->ChosenQPrior[j] = (enum qtype_enum *)ECA_CALLOC( (size_t)(j+1), sizeof(enum qtype_enum));
		}
		PrintEasyGrepLine(B, seed1, seed2);
		return(0);
	}
	
	
	
	/********************************************/
	
	
	
	// insert some values which we will pass as options later:
	B->AH_WindowWidth = 10;
	B->TM_WindowWidth = 10;


	FillAllMissingHoles(B);
	

	// if output data, then output data!
	if(OutPutData){
	  FILE * Datafile;
	  Datafile=fopen("DPinput.dat","w");
	  fprintf(Datafile,"Seqs: %d\n",B->NumSeq);
	  fprintf(Datafile,"Markers: %d\n",B->NumMarkers);

	  for(j=0;j<B->NumSeq;j++){
	    fprintf(Datafile,"DPIndiv%d",j);
	    for(i=0;i<B->NumMarkers;i++){
	      fprintf(Datafile,"\t%d",B->Y[j][i]);
	    }
	    fprintf(Datafile,"\n");;
	  }
	}

	// after doing this we are done with the random numbers, so record the seeds
	getsd(&seed1,&seed2);
	fprintf(LogFile,"\n\tEndingSeeds: %ld  %ld", seed1, seed2);
	fclose(LogFile);	
#ifndef RELEASE_VERSION
	SeedFile = fopen(SeedFileName,"w");
	fprintf(SeedFile,"%d %d",seed1,seed2);
	fclose(SeedFile);
#endif
	
	
	// Run this function to compute all the matrices.
	
	ComputeCodeLengths(B,B->Yobs);
	MoreEfficientFillInMatrices(B);
	
	// Now, B includes pointers to arrays that you should find useful:
	/*
		double **E;      -->   the matrix of entropies for the symbols denoting the different sequences in each block.
							    multiply this by the number of chromosomes in the data set to get the average cost in bits
							    of transmitting the data.
									
		double **CanonSeqMultParCost;	 -->  For the cost of sending the estimated parameter of the multinomial
											  model for cannonical seqs.
		double **SumOfCodeLengths;	-->  The sum of the code lengths for all the unique sequences in a block. 
			
		NOTE!  For a block starting at a and ending at b (a <= b) the corresponding quantity is found as, for example:
		
			B->E[b][a]     THE b IS THE FIRST SUBSCRIPT, EVEN THOUGH IT DENOTES THE END OF THE BLOCK.  
	
	*/



	if(nScoreBoundsSets>0){
	  // Open file
	  FILE *BoundsFile;
	  BoundsFile=fopen(ScoreBoundsFileName,"r");
	  
	  // Declare memory
	  nScoreBounds=(int *) ECA_CALLOC((size_t)nScoreBoundsSets,sizeof(int));
	  ScoreBoundsMatrix=(int **)ECA_CALLOC((size_t)nScoreBoundsSets,sizeof(int *));

	  // Read in Boundary matrix	  
	  for(i=0;i<nScoreBoundsSets;i++){
	    // Read in num bounds
	    fscanf(BoundsFile,"%d ",&nScoreBounds[i]);

	    // if nScoreBounds is reasonable Declare Memory 
	    if((nScoreBounds[i]>1)&&(nScoreBounds[i]<B->NumMarkers+2)){
	      ScoreBoundsMatrix[i]=(int *)ECA_CALLOC((size_t)nScoreBounds[i],sizeof(int));
	    
	      // Read in row of matrix
	      for(j=0;j<nScoreBounds[i];j++)
		fscanf(BoundsFile,"%d ",&ScoreBoundsMatrix[i][j]);
	    }
	  }

	  // Iterate over boundaries
	  for(i=0;i<nScoreBoundsSets;i++){
	    // if nScoreBounds is unreasonable, continue
	    if((nScoreBounds[i]<2)||(nScoreBounds[i]>B->NumMarkers+1)){
	      printf("\nNumber of bounds: %d, doesn't match data set.",nScoreBounds[i]);
	      continue;
	    }

	    gR=nScoreBounds[i]-1;
	     
	    printf("\nScoring the following boundary set:");
	    for(j=0;j<nScoreBounds[i];j++){
	      printf("%d ",ScoreBoundsMatrix[i][j]);
	    }

	    // exit if boundary endpoints exceeds data set
	    if((ScoreBoundsMatrix[i][nScoreBounds[i]-1]>B->NumMarkers)||(ScoreBoundsMatrix[i][0]<0)){
	      printf("\nBoundary set has end-points that are too large for this dataset of %d markers.",B->NumMarkers);
	      continue;
	    }
	    
	    // Warn the user if they're going to get a partial description length
	    if((ScoreBoundsMatrix[i][0]>0)||(ScoreBoundsMatrix[i][nScoreBounds[i]-1]<B->NumMarkers)){
	      printf("\nNote: The boundary set spans only a subset of the data!!");
	    }
	   
	    
	    printf("\nScore Bounds Score = %.5f\n", ScoreBounds(ScoreBoundsMatrix[i],nScoreBounds[i],B));
	    printf("\nOutputting detailed score summary to CodeLengths.tex\n");
	    PrintCostsLatex(B, ScoreBoundsMatrix[i],nScoreBounds[i], "CodeLengths.tex");	
	    
	  }
	  fclose(BoundsFile);
	  return(0);
	}


	gDP_NumBoundsInferred = 0;
	gR = 1;

	// Rvisited and Rscores are two arrays that will be used to
	// help avoid cycles (and in turn a lack of convergence) in
	// the iterative procedure for choosing gR below.  The i'th
	// element of Rvisited will be non-zero if gR has been set to
	// i+1 in the iterative algorithm Specifically the i'th
	// element will contain the inferred R for when gR was set to
	// i+1. The i'th element of Rscore will contain the score for
	// when gR=i+1. Cycles will be caught when an element of
	// Rvisited is non-zero when we attempt to write to it.

	Rvisited=(int*)ECA_CALLOC((size_t)B->NumMarkers,sizeof(int));
	Rscores=(double*)ECA_CALLOC((size_t)B->NumMarkers,sizeof(int));

	/* Herin lies the code for the iterative optimazition of gR */
	for(gR=0,gDP_NumBoundsInferred=2; gR != gDP_NumBoundsInferred-1;) {
	  
	  gR = gDP_NumBoundsInferred-1;
	  
	  if(gGreedy)
	    GreedyBlocks(B);
	  else
	    DP_Blocks(B);
	  	 
	  if(gR != gDP_NumBoundsInferred-1)  
	    printf("\nAssumed R = %d, Calculated R = %d ", gR,gDP_NumBoundsInferred-1);
	  else{
	    printf("\nAssumed and Calculated R values are converged to %d\n\n", gR);
	    break;
	  }			                  

	  if(Rvisited[gDP_NumBoundsInferred-1]>0){  // We're about to enter a cycle
	    Rvisited[gR]=gDP_NumBoundsInferred-1;
	    Rscores[gR]=gDP_BitScore;
	    gR=gDP_NumBoundsInferred-1;
	    ResolveCycle(Rvisited,Rscores,B);		  
	    break;
	  }else{ 
	    printf("\nContinuing iterations...");
	    Rvisited[gR]=gDP_NumBoundsInferred-1;
	    Rscores[gR]=gDP_BitScore;

	  }
	  


	} // for gR loop
	

	//SimAnnBlocks(B,B->NumSeq,1000,1000,1);
	JackBounds(B);
	
      	
      	
    /* print out the code lengths report  */
	printf("Outputting detailed score summary to CodeLengths.tex\n");
	PrintCostsLatex(B, gDP_Inferred_Bounds,gDP_NumBoundsInferred, "CodeLengths.tex");
      	      	
	PrintEasyGrepLine(B, seed1, seed2);
/*	printf("GR: %d\n",gR);
	
	printf("\n\nStored Values in the global array:");
	printf("\ngDP_NumBoundsInferred = %d", gDP_NumBoundsInferred);
	printf("\ngDP_Inferred_Bounds = (");
	for(i=0;i<gDP_NumBoundsInferred;i++)  {
		printf("%d  ",gDP_Inferred_Bounds[i]);
	}
	printf("\n");
*/
	
	/* Print out the result of ScoreBounds */
/*	printf("\nScore Bounds Score = %.5f\n", ScoreBounds(gDP_Inferred_Bounds,gDP_NumBoundsInferred,B)); */

#ifndef RELEASE_VERSION
	// now print out the canonical seqs observed:
	sprintf(gCanonSeqFileName,"InferredCanonSeqs.txt");
	tempfile = fopen(gCanonSeqFileName,"w");  // open and close to wipe a clean slate on it
	fclose(tempfile);
	for(i=1;i<gDP_NumBoundsInferred;i++)  {
		MoreEfficientEntropy(gDP_Inferred_Bounds[i-1],B,gDP_Inferred_Bounds[i],0);
	}
#endif

	
	/*
	sprintf(gCanonSeqFileName,"DalyCanonSeqs.txt");
	tempfile = fopen(gCanonSeqFileName,"w");  // open and close to wipe a clean slate on it
	fclose(tempfile);
	for(i=1;i<DalyNumBounds;i++)  {
		MoreEfficientEntropy(DalyBounds[i-1],B,DalyBounds[i],0);
	}
	
	scanf("%d",&i);
	
	GreedyFindByMDL(B);	
	printf("@%f ",ModelBitsAndDataBits(B->NumBounds, B->Bounds, B));	       
	for(i=0;i<B->NumBounds;i++) {
	  printf("%d ",B->Bounds[i]);	  
	}
	putchar('\n');
	*/
	// record the finishing time.
	StartTime = time(NULL);
	LogFile = fopen(LogFileName,"a");
	fprintf(LogFile, "\n\tProgram %s Completed %s  ",argv[0], ctime(&StartTime));
	fclose(LogFile);
	
	
	exit(0);	
	

}




void PrintHeterozygosities(MDL_struct *B)
{
	int i,j;
	double pq[2];
	double hz=0.0, ave_hz=0.0, tot, m;
	int idx, *hist=(int *)ECA_CALLOC(15, sizeof(int));
	
	for(i=0;i<B->NumMarkers;i++)  {
		pq[0]=0.0; pq[1]=0.0;
		for(j=0;j<B->NumSeq;j++)  {
			if(B->Yobs[j][i]>=0)
				pq[B->Yobs[j][i]] += 1.0;
			else if(B->Yobs[j][i]==AMBIG_HET)  {
				pq[0] += .5;
				pq[1] += .5;
			}
		}
		tot = pq[0] + pq[1];
		hz = 2.0 * (pq[0]/tot) * (pq[1]/tot);
		printf("\nMarker %d:   %f  Counts: %.1f  %.1f",i+1,hz,pq[0],pq[1]);
		ave_hz += hz;
		
		/* now do some stuff to look at the allele freq spec */
		if(pq[0]<pq[1]) 
			m=pq[0];
		else 
			m=pq[1];
		
		/* histogram of eleven bins: [0,.05), [.05,.1), ... , [.45,.5), [>=.5]  */
		idx = (int) ( (m/tot) * 20 );
		hist[idx]++;
		
	}
	printf("\n------\nAverage = %f",ave_hz/(double)B->NumMarkers);
	
	printf("\n\n\nAllele freq histograms:");
	for(i=0;i<12;i++)  {
		printf("\n[%.2f,%.2f)  %d  %f",  (double)i/20.0, (double)(i+1.0)/20.0, hist[i], (double)hist[i]/B->NumMarkers);
	}
	
	free(hist);
}


/* just prints out the output line to the buffer that we can grep out */
void PrintEasyGrepLine(MDL_struct *B, long seed1, long seed2)
{
	int i;
	
	printf("\nThe following line is designed to be easy to grep (grep using a backslash-escaped asterix) out of the stdout.\n");
	printf("\n*%f,%d,",gDP_BitScore,gDP_NumBoundsInferred);
	for(i=0;i<gDP_NumBoundsInferred;i++){
	  printf("%d ",gDP_Inferred_Bounds[i]);
	}
	printf(",");
	for(i=1;i<gDP_NumBoundsInferred-1;i++){
	  printf("%.3f ",gKLDs[i]);
	}
	printf(",");
	for(i=0;i<gDP_NumBoundsInferred-1;i++)  {  int a; int b;
		a = gDP_Inferred_Bounds[i];
		b = gDP_Inferred_Bounds[i+1] - 1;
		printf(" %d", ShiftQPrior((int)B->ChosenQPrior[b][a]));
	}
	printf(",%ld,%ld\n",seed1,seed2);
	
}

	
void usage(){
  fprintf(stderr,"usage: DPblocks <Input file> <Options>\n");
  fprintf(stderr,"\tOptions:\n");
  fprintf(stderr,"\t\t-g   :   To use the IADP algorithm.  This is faster\n"); 
  fprintf(stderr,"\t\t         than the default IDP algorithm. \n\n");  
  fprintf(stderr,"\t\t-s <seed 1> <seed 2> : To manually enter seeds.\n");
  fprintf(stderr,"\t\t                       For example: -s 234876 3287623\n");
  fprintf(stderr,"\t\t                       If absent, seeds are generated according\n");
  fprintf(stderr,"\t\t                       to the current time\n");
#ifndef RELEASE_VERSION
  fprintf(stderr,"\t\t-v <integer>         : To control verbosity of output.  \n"); 
  fprintf(stderr,"\t\t-l <Log file name>   : To choose log file name \n");
  fprintf(stderr,"\t\t-b <Seed file name>  : To read seeds from a file \n");
  fprintf(stderr,"\t\t-c                   : To use Wang's (2002) method\n");  
  fprintf(stderr,"\t\t-h                   : To print out the site by site heterozygosities, then quit\n");  
  fprintf(stderr,"\t\t-q<integer or char>  : To choose q cost function\n");
  fprintf(stderr,"\t\t                       i.e. -qc, -qe, -qo, -qm equivalent to\n");
  fprintf(stderr,"\t\t                            -q1, -q2, -q3, -q0\n");
  fprintf(stderr,"\t\t-B <integer> <file name> : To choose a bounds file and specify number of sets in the file.\n");
  fprintf(stderr,"\t\t-O : Output the filled in dataset to DPinput.dat\n");
  fprintf(stderr,"\t\t-t <double>          : Do Wang's method but require at least a proportion  \n");
  fprintf(stderr,"\t\t                       <double> (which should be between 0.0 and .25, non-inclusive)\n");
  fprintf(stderr,"\t\t                       of the chromosomes to have the least frequent haplotype category\n");
#endif  // RELEASE_VERSION
	
  fprintf(stderr,"\n\n\nAn example command line:");
  fprintf(stderr,"\n\t DP_Blocks MyDataFile.txt -g -s 234 876\n");
  
  
  exit(1);

}

struct DP_elem{
  double Val;
  int r,c;
  int LastBound;
  struct DP_elem *Point;
};


void DP_Blocks(MDL_struct *S) {
  struct DP_elem **M;
  int i,j,k;
  double t;
  int nSites;
  double MinCost;
  int MinSite;
      
  double BoundsCost(int LastBoundPos,int nSites);
  void TraceBack(struct DP_elem *P);
  double IntervalCost(int z,int a,int b,MDL_struct *B);


  nSites=S->NumMarkers;
  
  // Declare Memory for the DP matrix
  M=(struct DP_elem **)calloc((size_t)(nSites+1),sizeof(struct DP_elem *));
  if(!M) 
      fprintf(stderr,"Mem alloc error in DP_Blocks.");  
  for(i=0;i<nSites+1;i++){
    M[i]=(struct DP_elem *)calloc((size_t)(nSites+1),sizeof(struct DP_elem));
    if(!M[i]) 
      fprintf(stderr,"Mem alloc error in DP_Blocks.");
  }

  // Initialize r,c info.  Used by Traceback function
  for(i=0;i<nSites+1;i++){
    for(j=i;j<nSites+1;j++){
      M[i][j].r=i;
      M[i][j].c=j;
    }
  }

  // Initialize 0,0 entry  
  M[0][0].Point=0;
  // Set the initial score to the number bits needed to communicate the number of bounds
  M[0][0].Val=(log(nSites))/(log(2.0));


  // Initialize first row
  for(i=1;i<nSites+1;i++) {
    //cost to communicate num bounds+ cost to communicate the first interval
    M[0][i].Val=M[0][0].Val+IntervalCost(0,0,i-1,S);
    //And cost to specify boundary
    if(i<nSites)
      M[0][i].Val+=BoundsCost(0,nSites);
    M[0][i].Point=&M[0][0];
    //printf("\n0 %d %lf",i,M[0][i].Val);
  }
  
  // Complete the matrix
  for(j=1;j<nSites;j++) {
    if(gVerboseFlag>0)
      printf("\nComputing DP matrix row: %d",j);
    // Fill in row
    for(k=j+1;k<nSites+1;k++) {

      // Initialize MinCost 
      MinCost=M[0][j].Val+IntervalCost(0,j,k-1,S);
      if(k<nSites)
	MinCost+=BoundsCost(j,nSites);
      MinSite=0;

      // Minimize 
      for(i=1;i<j;i++){
	//printf("\nOn: %d, %d, %d",j,k,i);
	t=M[i][j].Val+IntervalCost(i,j,k-1,S);
	if(k<nSites)
	  t+=BoundsCost(j,nSites);	

	if(t==MinCost){
	  printf("\nTie: i=%d j=%d t=%f MinCost=%f MinSite=%d. Choosing solution that gives longer blocks.",i,j,t,MinCost,MinSite);
	  // Because i starts at 0, not updating the MinCost and MinSite will 
	  // have the effect of creating longer blocks, and in the end
	  // less blocks (although I haven't proven this).
	}else if(t<MinCost){
	  //printf("\nMin: %d %f",i,t);
	  MinCost=t;
	  MinSite=i;
	}
      }
      M[j][k].Val=MinCost;
      M[j][k].Point=&M[MinSite][j];
    }
  }


  // ECA ADDED SOME STUFF HERE TO STORE THE BOUNDARIES IN a global 
  // array int *gDP_Inferred_Bounds with length gDP_NumBoundsInferred
  gDP_NumBoundsInferred = 0;
  gDP_Inferred_Bounds = (int *)ECA_CALLOC( (size_t)nSites, sizeof(int));
  
  // Minimize over last blocks

  // Initialize MinCost MinSite
  MinCost=M[0][nSites].Val;
  MinSite=0;

  for(j=1;j<nSites;j++){
    if(M[j][nSites].Val<MinCost){
      MinCost=M[j][nSites].Val;
      MinSite=j;
    }
  }

  gDP_BitScore=MinCost;

  TraceBack(&M[MinSite][nSites]);



  // Print S matrix
  if(gVerboseFlag>1){
  printf("\nE\n");
  for(i=0;i<nSites;i++) {
    printf("      %2d",i+1);
  }
  printf("\n");  
  for(i=0;i<nSites;i++) {
    printf("%2d",i);
    for(j=0;j<i;j++) {
      printf("        ");
    }
    for(j=i;j<nSites;j++) {
      printf("%7.2f ", S->E[j][i]);
    }
    printf("\n");
  }
	
  printf("\nSumOfCodeLengths\n");
  for(i=0;i<nSites;i++) {
    printf("      %2d",i+1);
  }
  printf("\n");  
  for(i=0;i<nSites;i++) {
    printf("%2d",i);
    for(j=0;j<i;j++) {
      printf("        ");
    }
    for(j=i;j<nSites;j++) {
      printf("%7.2f ", S->SumOfCodeLengths[j][i]);
    }
    printf("\n");
  }
	
  printf("\nNT\n");
  for(i=0;i<nSites;i++) {
    printf("      %2d",i+1);
  }
  printf("\n");  
  for(i=0;i<nSites;i++) {
    printf("%2d",i);
    for(j=0;j<i;j++) {
      printf("        ");
    }
    for(j=i;j<nSites;j++) {
      printf("%7d ", S->NT[j][i]);
    }
    printf("\n");
  }


  // print bounds cost 
  printf("\n");
  for(i=0;i<nSites;i++) {
    printf("    %2.2f",BoundsCost(i,nSites));
  }

  // print DP matrix
  printf("\nDP Matrix\n");
  for(i=0;i<nSites+1;i++) {
    printf("\t%d",i);
  }
  printf("\n");  
  for(i=0;i<nSites+1;i++) {
    printf("%d",i);
    for(j=0;j<i;j++) {
      printf("\t");
    }
    for(j=i;j<nSites+1;j++) {
      printf("\t%.1f", M[i][j].Val);
    }
    printf("\n");
  }
  }

  //putchar('\n');
  
  // ECA added freeing the memory to M here
  for(i=0;i<nSites+1;i++){
    free(M[i]);
  }
  free(M);
}

void TraceBack(struct DP_elem *P){

  if(P->Point) {
    TraceBack(P->Point);
  }

  gDP_Inferred_Bounds[gDP_NumBoundsInferred++] = P->c;  // ECA added this line

}

double IntervalCost(int z,int a,int b,MDL_struct *B) {

  double value=0.0;
  double DeltaCost, DataCost, PCost, log2NabSum, FixedBitCost, KLD;
  int numdelts;
  // double CostWithMixture(int z, int a, intb, MDL_struct *B);
  
  // add to it the part for specifying the blocks
  value += B->SumOfCodeLengths[b][a];
  
  // add to it the part for specifying the frequency distribution of
  // the canonical sequences, in other words:
  // the part for encoding the multinomial model of proportions
  value += B->CanonSeqMultParCost[b][a];
  
  if(gQPrior==MIXTURE)   /* add the spare bits for specifying the type of prior dsn on q */
  	value += log(NUM_QPRIOR_DSNS)/log(2.0);
  
  // Add remaining costs... these include cost of transition probablities 
  // as well as costs of data.
  if(a==0)
    value += (double)B->NumSeq*B->E[b][a];
  else
    value += CostWithMixture(z,a,b,B,&DataCost, &DeltaCost, &PCost,&numdelts, &log2NabSum, &FixedBitCost, &KLD);

  
  return(value);
  

}

double BoundsCost(int LastBoundPos, int nSites){

  double ReturnVal=0.0;
  double CostSum=0.0;
  int i=0;

  for(i=1;i<(gR+1);i++) {
    CostSum+=log(i)/log(2.0);
  }
  if(gR>1)
    CostSum/=(gR-1);
  ReturnVal=log(nSites)/log(2.0)-CostSum;

  // Assumed R version:
  return(ReturnVal);

  // Relative Positions Version...
  // The cost adding a bound is log base 2 of nSites-LastBoundPos
  //return((log(nSites-LastBoundPos))/(log(2.0)));

  // Absolute Positions Version...
  // Note: Function prototype has to change
  // The cost of adding the nth bound is log base 2 of nSites-nBounds
  // return(log(nSites-nBounds)/log(2.0));  
}


// compute d-prime between sites a b.
double pwise_d_prime(MDL_struct *B, int a, int b, int *ssize)
{
	int i;
	double Dab, Dmax, xab=0.0,nab=0.0, na=0.0, nb=0.0, pa=0.0, pb = 0.0;
	
	// compute xab.  Do it for the 1-1 homozygotes
	for(i=0;i<B->NumSeq;i++)  {
		if(B->Yobs[i][a]>=0 && B->Yobs[i][b]>=0)  {  // if they both have no missing data
			nab += 1.0;
			na += 1.0;
			nb += 1.0;
			if(B->Yobs[i][b]==1)
				pb += 1.0;
			if(B->Yobs[i][a]==1)
				pa += 1.0;
			if(B->Yobs[i][a]==1 && B->Yobs[i][b]==1)
				xab += 1.0;
		}
	}
	xab /= nab;
	pa /= na;
	pb /= nb;
	
	Dab = xab - pa * pb;
	
	if(Dab<0) {
		Dmax = (pa * pb) < ( (1.-pa)*(1.-pb) ) ? (pa * pb)  :  ( (1.-pa)*(1.-pb) );
	}
	else {
		Dmax = (pa*(1.-pb)) < (pb*(1.-pa) )  ?  (pa*(1.-pb)) : (pb*(1.-pa) );
	}
	
	*ssize = nab; 
	
	return(Dab/Dmax);
}

/*
	Goes through the pairwise haplotypes at markers a and b.  Returns a 
	1 if it passes the four gamete test and a 0 otherwise.
	
	Loci at which there is truly missing data don't get used, and loci with
	unresolved heterozygosity don't get used either.  
*/
int PassesTheFourGameteTest(MDL_struct *B, int a, int b)
{
	int i,x[2][2], ret = 0;
	
	x[0][0] = 0;
	x[0][1] = 0;
	x[1][0] = 0;
	x[1][1] = 0;
	
	for(i=0;i<B->NumSeq;i++)  {
		if(B->Yobs[i][a]>=0 && B->Yobs[i][b]>=0)  { 
			x[B->Yobs[i][a]] [B->Yobs[i][b]] += 1;
		}
	}
	
	if(x[0][0] * x[0][1] * x[1][0] * x[1][1] == 0)
		ret = 1;
	
	return(ret);
	
}



/*
	Goes through the pairwise haplotypes at markers a and b.  Returns a 
	0 if only three or fewer of the haplotypes are observed, and it 
	returns n otherwise, where n is the number of haplotypes belonging to
	the smallest of the four classes.
	
	Loci at which there is truly missing data don't get used, and loci with
	unresolved heterozygosity don't get used either.  
*/
int CountIncompatibilities(MDL_struct *B, int a, int b, int *NumNoMissing)
{
	int i,j,x[2][2], low=B->NumSeq;
	
	x[0][0] = 0;
	x[0][1] = 0;
	x[1][0] = 0;
	x[1][1] = 0;
	
	*NumNoMissing = 0;
	
	for(i=0;i<B->NumSeq;i++)  {
		if(B->Yobs[i][a]>=0 && B->Yobs[i][b]>=0)  { 
			x[B->Yobs[i][a]] [B->Yobs[i][b]] += 1;
			(*NumNoMissing) += 1;
		}
	}
	
	if(x[0][0] * x[0][1] * x[1][0] * x[1][1] == 0)
		return(0);
		
	else  {
		for(i=0;i<2;i++)  {
			for(j=0;j<2;j++)  {
				if(x[i][j] < low)
					low = x[i][j];
			}
		}
	}
	return(low);
	
	
}


/* this is like DoWang, except it only counts things as an "incompatibility"
if the proportion of incompatible pairs (out of all those that don't have missing
data) is greater than slop */ 
void DoWangWithSlop(MDL_struct *B, double slop)
{
	int a,b, NumInc, NumNoMiss;
	
	// ECA ADDED SOME STUFF HERE TO STORE THE BOUNDARIES IN a global 
	// array int *gDP_Inferred_Bounds with length gDP_NumBoundsInferred
	gDP_NumBoundsInferred = 0;
	gDP_Inferred_Bounds = (int *)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(int));
	gDP_Inferred_Bounds[0] = 0;  /* we know the first is always 0 */
	a=b=0;
	while(b<B->NumMarkers-1)  {  /* work through, sequentially applying the FGT */
		b++;
		NumInc = CountIncompatibilities(B,a,b,&NumNoMiss);
		if((double)NumInc/(double)NumNoMiss > slop) {
			gDP_Inferred_Bounds[++gDP_NumBoundsInferred] = b;
			a = b;
		}
		
	}
	/* at the end, we have to put the last one in there  */
	gDP_Inferred_Bounds[++gDP_NumBoundsInferred] = B->NumMarkers;
	/* JN: And we need to increment the counter one more time */
	gDP_NumBoundsInferred++;
}



void DoWang(MDL_struct *B)
{
	int a,b;
	
	// ECA ADDED SOME STUFF HERE TO STORE THE BOUNDARIES IN a global 
	// array int *gDP_Inferred_Bounds with length gDP_NumBoundsInferred
	gDP_NumBoundsInferred = 0;
	gDP_Inferred_Bounds = (int *)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(int));
	gDP_Inferred_Bounds[0] = 0;  /* we know the first is always 0 */
	a=b=0;
	while(b<B->NumMarkers-1)  {  /* work through, sequentially applying the FGT */
		b++;
		if(!SeqPairsPassFGT(B,a,b)) {
			gDP_Inferred_Bounds[++gDP_NumBoundsInferred] = b;
			a = b;
		}
		
	}
	/* at the end, we have to put the last one in there  */
	gDP_Inferred_Bounds[++gDP_NumBoundsInferred] = B->NumMarkers;
	/* JN: And we need to increment the counter one more time */
	gDP_NumBoundsInferred++;
}


/*  
	Applies the FGT between markers a and b, a+1 and b, a+2 and b, and so forth,
	up to b-1 and b.  Returns a one if they all pass the FGT.  If not it returns a zero.
	Assumes that b >= a.
*/
int SeqPairsPassFGT(MDL_struct *B,int a,int b)
{
	int aa, the_result=1;
	
	if(a>b) {
		fprintf(stderr,"\nError!  b < a in SeqPairsPassFGT()!!\n\nExiting to system...\n\n");
		exit(1);
	}
	for(aa=a;aa<b;aa++)  {
		if(!PassesTheFourGameteTest(B,aa,b)) {
			the_result = 0;
			break;
		}
	}
	
	return(the_result);
}

void GreedyBlocks(MDL_struct *S) {
  struct DP_elem *M;
  int i,j;
  double t;
  int nSites;
  double MinCost;
  int MinSite;
  double BoundsCost(int LastBoundPos,int nSites);
  void GreedyTraceBack(struct DP_elem *P);
  double IntervalCost(int z,int a,int b,MDL_struct *B);

  nSites=S->NumMarkers;
  
  // Declare Memory for the DP matrix
  M=(struct DP_elem *)calloc((size_t)(nSites+1),sizeof(struct DP_elem));
  if(!M) 
      fprintf(stderr,"Mem alloc error in DP_Blocks.");  

  // Initialize c to the vector element id.  Used by Traceback function
  for(i=0;i<nSites+1;i++)
      M[i].c=i;

  // Initialize 0 entry  
  M[0].Point=0;
  // Set the initial score to the number bits needed to communicate the number of bounds
  M[0].Val=(log(nSites))/(log(2.0));
  M[1].Val=M[0].Val+IntervalCost(0,0,0,S);
  M[1].Point=&M[0];
  M[0].LastBound=M[1].LastBound=0;

  // Complete the vector
  for(i=2;i<nSites+1;i++) {
    //printf("\nPerforming approximate DP optimization.  Currently at site: %d",i);
    // Initialize MinCost,MinSite
    MinCost=M[0].Val+IntervalCost(0,0,i-1,S);
    if(i<nSites)
      MinCost+=BoundsCost(0,nSites);
    // Minimize
    MinSite=0;
    //printf("\nOn: %d ",i);
    for(j=1;j<i;j++){
      //printf("\nOn: %d, %d Val: %f IC: %f",i,j,M[j].Val,IntervalCost(M[i].LastBound,j,i-1,S));
      t=M[j].Val+IntervalCost(M[j].LastBound,j,i-1,S);
      if(i<nSites)
	t+=BoundsCost(j,nSites);
      if(t==MinCost){
	printf("\nTie: %d %d.  Choosing solution that gives longer blocks.  ",i,j);
	// Because we begin at j=0 and decrease block length from
	// there, by not updating on the tie, we keep the longer
	// block.
      }else if(t<MinCost){
	//printf("\nMin: %d ",j);
	MinSite=j;
	MinCost=t;
      }
    }

    // Set new element
    M[i].Point=&M[MinSite];
    M[i].Val=MinCost;
    M[i].LastBound=MinSite;
    if(gVerboseFlag>3)
      printf("%f %d",M[i].Val,M[i].LastBound);
  }
  
  
  // ECA ADDED SOME STUFF HERE TO STORE THE BOUNDARIES IN a global 
  // array int *gDP_Inferred_Bounds with length gDP_NumBoundsInferred
  gDP_NumBoundsInferred = 0;
  gDP_Inferred_Bounds = (int *)ECA_CALLOC( (size_t)nSites, sizeof(int));
  
  // Minimize over last blocks
  gDP_BitScore=M[nSites].Val;

  TraceBack(&M[nSites]);
  free(M);
}

void JackBounds(MDL_struct *S){


  int* Bounds;
  int nSites;
  double OrigScore;
  nSites=S->NumMarkers;
  // Make a temporary boundaries array
  Bounds = (int *)ECA_CALLOC( (size_t)gDP_NumBoundsInferred, sizeof(int));
  gJackknifeScores = (double *)ECA_CALLOC ((size_t)gDP_NumBoundsInferred,sizeof(double));

  Bounds[0]=0;
  OrigScore=ScoreBounds(gDP_Inferred_Bounds,gDP_NumBoundsInferred,S);

#ifndef RELEASE_VERSION
  if(gVerboseFlag>0)
    printf("\nOriginal Score: %f",OrigScore);

  
  // For each internal boundary 
  for(i=1,j=1;i<gDP_NumBoundsInferred-1;i++,j=1) {
    
    

    // Assign to Bounds, but skip dropped the boundary
    for(k=1;k<gDP_NumBoundsInferred-2;k++) {
      if(j==i)
	j++;
      Bounds[k]=gDP_Inferred_Bounds[j];
      j++;
    }

    // Set last boundary
    Bounds[gDP_NumBoundsInferred-2]=gDP_Inferred_Bounds[gDP_NumBoundsInferred-1];

    // Rescore
    gJackknifeScores[i]=ScoreBounds(Bounds,gDP_NumBoundsInferred-1,S);
    
    // Store
    if(gVerboseFlag>0)
      printf("\nBound dropped: %d Score: %f Score ratio: %f",gDP_Inferred_Bounds[i],gJackknifeScores[i],gJackknifeScores[i]/OrigScore);
  }
  free(Bounds);
#endif

}

double ScoreBounds(int* Bounds,int nBounds,MDL_struct *S){ 

  double Score=0.0;
  int nSites;
  int LastBound=0;
  int i;

  nSites=S->NumMarkers;
  Score=(log(nSites))/(log(2.0));
  for(i=1;i<nBounds;i++) {
    Score+=IntervalCost(LastBound,Bounds[i-1],Bounds[i]-1,S);
    if(i<nBounds-1)
      Score+=BoundsCost(Bounds[i-1],nSites);
    //printf("\nBound %d Score: %f",Bounds[i],Score);
    LastBound=Bounds[i-1];
  }

  return(Score);
}

/* This function takes a boundaries array in binary format, converts it to */
/* the difference format and scores it with ScoreBounds */
double ScoreBinaryBounds(int* BinBounds,MDL_struct *S){
  double ScoreBounds(int* Bounds,int nBounds,MDL_struct *S);
  
  int* Bounds;
  int i,j;
  int nSites;
  int LastBound;
  double Score;
  nSites=S->NumMarkers;
  Bounds=(int*)ECA_CALLOC((size_t)nSites,sizeof(int));

  LastBound=0;
  for(i=0,j=0;i<nSites;i++) {
    if(BinBounds[i]==1){
      Bounds[j]=i-LastBound;
      j++;
    }
  }
  Score=ScoreBounds(Bounds,j,S);
  free(Bounds);
  return(Score);
  
}


/* this properly prints out the Q prior used, so that it is the same as in the paper */
int ShiftQPrior(int a) {
	switch(a) {
		case(1):
			return(1);
			break;
		case(2):
			return(3);
			break;
		case(3):
			return(2);
			break;
		default:
			return(0);
	}
}

/*  For a given set of bounds, print out a detailed report in LaTeX tabular format 
	that summarizes all the costs and things for a given set of bounds.  This can only
	be done after the entropy matrices have been filled out and computed, etc.
	
	Note that NumBounds in this case includes 0 and NumMarkers
	
	This also puts the KL divergences into a global array that can be accessed later
	to print stuff to a buffer for analysis later.
*/
void PrintCostsLatex(MDL_struct *B, int *Bounds, int NumBounds, char *FileName)
{
	double 	Ak=0.0, Ak_k,   /*  the code length for specifying the types (total and in row k) */
			q=0.0,	q_k,	/* code length for the multinomial proportions */
			s=0.0,  s_k,  	/* code length for the number of types in a block */
			x=0.0,  x_k,   /* code length for transmitting the X's (the data) */
			delta=0.0, delta_k,  /* code length for specifying the position of the one's in the delta-matrix */
			p=0.0,  p_k,  	/* cost of the specified transition probs in the matrix P */
			L=0.0,		/* code length for L, the lengths of the blocks */
			R = log(B->NumMarkers)/log(2.0),  	/* the constant cost term that gets added in to initialize the DP */
			BlockCost,		/* the total cost for all the parts in a single block  */
			MixtureCost = 0.0, MixtureCost_k, TotalCost,  /* MixtureCost is just to compute the total by what CostWithMixture returns, rather than
											   using its component parts and summing them together later */
			NumOnes = 0.0, NumOnes_k,  /* for the code length of specifying the number of 1's in the matrix Delta */
			KLD_k;  /* for getting the Kullback-Leibler Divergence between P* and q at each block */
	int numDelts=0, numDelts_k; /* The number of deltas = 1 in Delta, i.e. the number of trans probs specified for the row */
	int i,k,a,b,z;
	int S, Sprevious = 0;  /* the number of canonical seqs at this block and at the last one */
	FILE *out;
	char boing;
	double log2NabSum=0.0, FixedBitCost=0.0, log2NabSum_k, FixedBitCost_k, MixtureQPriorBits_k,MixtureQPriorBits=0.0;

	
	
	while( (out=fopen(FileName, "w"))==NULL)  {
		fprintf(stderr, "\n\nFailed to open file %s, in PrintCostsLatex.\n\n",FileName);
		fprintf(stderr, "You  probably have that file open in another application\n");
		fprintf(stderr, "Please close it and then hit return\n\n");
		scanf("%c",&boing);
	}
	
	
	/*  allocate memory to the global KL divergence array */
	gKLDs = (double *)ECA_CALLOC( (size_t)NumBounds-1, sizeof(double));
	
	
	/* print the first opening LaTeX stuff */
	fprintf(out, "\\documentclass[]{article}\n");
	fprintf(out,"\n\\def\\bmstyle#1#2{\\mbox{\\boldmath$#1#2$}}\\def\\bm#1{\\mathpalette\\bmstyle{#1}}\n\\newcommand{\\vecX}{\\bm{X}}\\newcommand{\\vecq}{\\bm{q}}");
	fprintf(out,"\n\\newcommand{\\matP}{\\mathbf{P}}\\newcommand{\\matDelta}{\\bm{\\Delta}}\n\\newcommand{\\Seqset}{\\mathcal{A}}\\newcommand{\\CL}{\\varphi}");
	fprintf(out,"\n\\newcommand{\\vecE}{\\bm{E}}");
	fprintf(out,"\n\\newcommand{\\vech}{\\bm{h}}");
	fprintf(out,"\n\\newcommand{\\vecZ}{\\bm{Z}}");
	fprintf(out,"\n\n\\setlength{\\oddsidemargin}{-.84in}\n\\setlength{\\textwidth}{10in}");
	fprintf(out,"\n\\begin{document}\n\n");
	fprintf(out,"\n\n\\small\n\n");
	fprintf(out,"\n\\begin{center}{\\LARGE Output of Program \\texttt{MDBlocks} \\\\ {\\normalsize(written by Eric C. Anderson and John Novembre)}}\\end{center}");
	fprintf(out,"\n\nDataset: \\texttt{%s}\n\n",B->InfileName);
	fprintf(out,"Number of Sequences: $N=%d$\n\n",B->NumSeq);
	fprintf(out,"Number of Markers: $M=%d$\n\n",B->NumMarkers);
	fprintf(out,"``Prior For $\\vecq$\" = {\\sl %s}\n\n",qtype2str(gQPrior));
	fprintf(out,"Inferred Number of Blocks: $R=%d$\n\n",NumBounds-1);
	fprintf(out,"\n\\vspace*{.25in}\n");
	fprintf(out,"\n\n\\begin{tabular}{lcccccc|ccccccccccc|r}");
	fprintf(out,"\n\\hline\\hline \\\\");
	fprintf(out,"\n$k$ & $a$ & $E^{(k)}$ & $S^{(k)}$ & $\\sum_rn_{r}^{(k)}$ & K-L & $I_{\\vecq}^{(k)}$ & $\\CL(R)$ & $\\CL(\\vecE)$ &");
	fprintf(out,"$\\CL(\\mathrm{FxBt})$  & $\\CL(\\mathrm{NumDelts})$  & $\\CL(S^{(k)})$ & $\\CL(\\Seqset^{(k)},\\vech^{(k)})$ & $\\CL(I_{\\vecq}^{(k)})$  &$\\CL(\\vecq^{(k)})$  & $\\CL(\\matDelta^{(k)})$ & $\\CL(\\matP^{(k)})$ & $\\CL(\\vecX^{(k)},\\vecZ^{(k)})$ ");
	fprintf(out," & Total \\\\ \\hline");	 


	/* compute the boundary cost (which is the same for every block) given gR */
	/* we just add it all up and report it as a total cost, rather than breaking it
		down by block  */
	L=0.0;
	for(i=1;i<(gR+1);i++) {
		L+=log(i)/log(2.0);
	}
	if(gR>1)
    	L/=(gR-1);
    L = log(B->NumMarkers)/log(2.0)-L;
    /* the lines above give the cost for each internal boundary.  Recall that there are
       NumBounds-2 such internal bounds.  So the total cost for this comes by multiplying
       L by (NumBounds-2) */
    
    L *= (double)(NumBounds-2);
    
    /* here is a cost that will be the same for all blocks: */
    if(gQPrior==MIXTURE)
	    MixtureQPriorBits_k = log(NUM_QPRIOR_DSNS)/log(2.0);
	else 
		MixtureQPriorBits_k = 0.0;
		
		

	for(k=0;k<NumBounds-1;k++)  {  /* note that k+1 will be the subscript of the block using 
									  base-1 subscripting as done in the manuscript */
		a = Bounds[k];
		b = Bounds[k+1]-1;
		
		Ak_k = B->SumOfCodeLengths[b][a];
		s_k = log(B->NumSeq) / log(2.0);
		q_k = B->CanonSeqMultParCost[b][a] - s_k;  /* the cost of the number of types has already
													   been included in CanonSeqMultParCost, so we
													   have to subtract it out  */
		S = B->Tr[b][a].N;
		
		

		
		if(k==0)  {  /* if we are looking at the first block */
			x_k = (double)B->NumSeq*B->E[b][a];  /* cost of the data  in the first block 
													as computed also in IntervalCost */
			delta_k = 0.0;
			p_k = 0.0;
			numDelts_k = 0;
			NumOnes_k = 0.0;
			MixtureCost_k = x_k;
			FixedBitCost_k = 0.0;
			log2NabSum_k = 0.0;
			KLD_k = 0.0;
		}
		
		else {  /*  otherwise we have to deal with the mixture stuff  */
			z = Bounds[k-1];
			MixtureCost_k = CostWithMixture(z,a,b,B,&x_k, &delta_k, &p_k,&numDelts_k,
								&log2NabSum_k, &FixedBitCost_k, &KLD_k);
			NumOnes_k = (double)Sprevious * ( log((double)S) / log(2.0) );
		}
		
		Sprevious = S;
		
		/* then we increment all the sums over blocks */
		Ak += Ak_k;
		q += q_k;
		s += s_k;
		x += x_k;
		delta += delta_k;
		p += p_k;
		numDelts += numDelts_k;
		NumOnes += NumOnes_k;
		FixedBitCost += FixedBitCost_k;
		log2NabSum += log2NabSum_k;
		MixtureQPriorBits += MixtureQPriorBits_k;
		
		gKLDs[k] = KLD_k;
		
		BlockCost = Ak_k + q_k + s_k + x_k + delta_k + p_k + FixedBitCost_k + log2NabSum_k, MixtureQPriorBits_k;
		MixtureCost += MixtureCost_k;
		
	
		
		/* Now we want to print out all this stuff as a row in a text file */
		/* first print the block boundaries and then the number of types in it */
		
		fprintf(out, "\\\\ \n %d & %d & %d & %d & %d & %.2f & %d & --- & --- &",k+1,a,b,S,numDelts_k, KLD_k, ShiftQPrior((int)B->ChosenQPrior[b][a]));
	
		
		/* then print out the code lengths for the various quantities */
		fprintf(out, "%.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f",
			           FixedBitCost_k, log2NabSum_k, s_k, Ak_k, MixtureQPriorBits_k,  q_k, delta_k, p_k,x_k,BlockCost);
		
		
		
		
	}
	
	TotalCost = Ak + q + s + x + delta + p + L + R + FixedBitCost + log2NabSum + MixtureQPriorBits;
	
	/* after all that we want to print the line of totals and close the file */
	/* Now we want to print out all this stuff as a row in a text file */

	/* print the block boundaries and then the number of types in it */
	fprintf(out, "\n \\\\ \\\\  \\hline \\\\ \n Tot & --- & --- & --- & --- & --- & --- & %.2f & %.2f &",R,L);

	/* then print out the code lengths for the various quantities */
	fprintf(out, "%.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f & %.2f",
		           FixedBitCost, log2NabSum,s, Ak, MixtureQPriorBits, q,delta, p,x, TotalCost);
	fprintf(out,"\n\\\\ \\hline \\hline \n\\end{tabular}");
	fprintf(out, "\n\n\\end{document}");
		           
	fclose(out);
}


char *qtype2str(enum qtype_enum Q)
{
	switch(Q) {
		case(MIXTURE):
			return("Mixture");
		case(STANDARD):
			return("Conditional Uniform");
		case(EWENS):
			return("Multivariate Ewens");
		case(OCCUPANCY):
			return("Equiprobable Occupancy");
		default:
			return("Unrecognized Q Prior");
	}
	
	return("Problem! Made it through switch() in qtype2str");
}
