
/*
Copyright (c). The Regents of the
University of California (Regents). All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby granted,
provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
for commercial licensing opportunities. Created by Eric C. Anderson and
John A. Novembre,
Department of Integrative Biology, University of California, Berkeley.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "ECA_MemAlloc.h"
#include "MathStatRand.h"

#include "MDL_Blocks.h"


/******************************************

MEMORY MANAGEMENT, ALLOCATION, DE-ALLOCATION

******************************************/

MDL_struct *AllocMDL_struct(int NumSeq, int NumMarkers){
	int i; 
	
	MDL_struct *B;
	
	// first create the MDL_struct
	B = (MDL_struct *)ECA_MALLOC(sizeof(MDL_struct));
	
	// then copy some values
	B->NumMarkers = NumMarkers;
	B->NumSeq = NumSeq;
	
	// make the E, NT, Node and Tips pointer point to NULL so we
	// will know to allocate memory to them later.
	B->E = NULL;
	B->CanonSeqMultParCost = NULL;
	B->ChosenQPrior = NULL;
	B->NT = NULL;
	B->Nodes = NULL;
	B->Tips = NULL;
	B->TipSeqs = NULL;
	B->CodeLengths = NULL;
	B->SumOfCodeLengths = NULL;
	B->TipsArray = NULL;
	B->TipsNums = NULL;
	B->TipsNums2 = NULL;
	B->S = NULL;
	B->Sets = NULL;
	B->Canon = NULL;
	B->FC = NULL;
	B->FCP = NULL;
	B->FCP_filled = 0;
	B->Tr = NULL;
	
	B->EntropyWeightFactor = 1.0;
	
	// then also allocate to Y and Yobs
	B->Y = (int **)ECA_CALLOC((size_t)B->NumSeq,sizeof(int *));
	B->Yobs = (int **)ECA_CALLOC((size_t)B->NumSeq,sizeof(int *));
	for(i=0;i<B->NumSeq;i++)  {
		B->Y[i] =(int *)ECA_CALLOC((size_t)B->NumMarkers,sizeof(int));
		B->Yobs[i] =(int *)ECA_CALLOC((size_t)B->NumMarkers,sizeof(int));
	}
	
	// and allocate to the Names
	B->Names = (char **)ECA_CALLOC((size_t)B->NumSeq,sizeof(char *));
	for(i=0;i<B->NumSeq;i++)  {
		B->Names[i] = (char *)ECA_CALLOC( (size_t)MAX_SEQUENCE_NAME_LENGTH, sizeof(char)); 
	}
	
	B->InfileName = (char *)ECA_CALLOC( (size_t)200, sizeof(char));
	
	return(B);
}


void DeAllocMissingDataInfo(missing_data_info *M)
{
	if(M->NSWM > 0)  {  // only free these if they were allocated to, which
						// would have happened only if NSWM was > 0
		free(M->M_Idx);
		free(M->NumMissingArray); 
		free(M->MissingIndices);
	}
	
	free(M);
}  

// allocate to a matrix of doubles in which  row i has i columns (base-0).
// the [b][a]-th element of this matrix will apply to the block
// extending from a to b inclusive.
double **AllocEntMat(int NumMarkers)
{
	int j;
	double **mat;
	
	// allocate to mat and to NT
	mat = (double **)ECA_CALLOC((size_t)NumMarkers,sizeof(double *));
	for(j=0;j<NumMarkers;j++)  {
		mat[j] = (double *)ECA_CALLOC( (size_t)(j+1), sizeof(double));
	}
	
	return(mat);
}

// allocate to a matrix of ints in which  row i has i columns (base-0).
// the [b][a]-th element of this matrix will apply to the block
// extending from a to b inclusive.
int **AllocNTMat(int NumMarkers)
{
	int j;
	int **NT;
	
	// allocate to NT
	NT = (int **)ECA_CALLOC((size_t)NumMarkers,sizeof(int *));
	for(j=0;j<NumMarkers;j++)  {
		NT[j] = (int *)ECA_CALLOC((size_t)NumMarkers,sizeof(int));
	}
	
	return(NT);
}





/******************************************

VARIABLE INITIALIZATION

******************************************/

// initialize all the values in the entropy matrix to -1.0's
void InitEntMat(double **mat, int NumMarkers)
{
	int a,b;
	
	for(a=0;a<NumMarkers;a++)  {
		for(b=a;b<NumMarkers;b++)  {
			mat[b][a] = -1.0;
		}
	}
}

// initialize all the values in the NT matrix to -1's
void InitNTMat(int **NT, int NumMarkers)
{
	int a,b;
	
	for(a=0;a<NumMarkers;a++)  {
		for(b=a;b<NumMarkers;b++)  {
			NT[b][a] = -1;
		}
	}
}

// initialize a node in the binary tree.  
void InitNode(struct bin_node *N)
{
	
	N->Visits = 0;
	N->Visits2 = 0;
	N->Idx = -1;
	N->Br[0] = NULL;
	N->Br[1] = NULL; 
	N->nOnes = 0;
	N->nZeroes = 0;

	N->SetPtr = NULL;
		
}




/******************************************

READING DATA IN

******************************************/

// reads in data from a file.  Allocates memory to
// an MDL struct and inserts the data in Yobs and
// makes sure things are initialized as they need to be in the MDL
// struct, then returns a pointer to the MDL_struct
MDL_struct *ReadSeqsAsBits(char *FileName)
{
	int i,j;
	FILE *in = fopen(FileName,"r");
	int NumSeq, NumMarkers;
	MDL_struct *B;
	char tempStr[500];
	
	sprintf(tempStr,"%s",FileName);
	while(in == NULL)  {
		printf("\nSorry my friend, file \"%s\" not found.\n",tempStr);
		printf("Please enter another file name\n->");
		scanf("%s",tempStr);
		in = fopen(tempStr,"r");
	}
	
	// scan the number of sequences and chromosomes
	fscanf(in," Seqs: %d\nMarkers: %d\n",&NumSeq, &NumMarkers);
	
	B = AllocMDL_struct(NumSeq,NumMarkers);
	
	// then scan in all the bits and the names as we go
	for(i=0;i<B->NumSeq;i++)  {
		fscanf(in," %s",B->Names[i]);
		if(gVerboseFlag)
			printf("%s ",B->Names[i]);
		for(j=0;j<B->NumMarkers;j++)  {
			fscanf(in," %d",&B->Yobs[i][j]);
		}
	}
	
	fclose(in);
	
	return(B);
}




/******************************************************

IMPLEMENTING THE MINIMUM DESCRIPTION LENGTH METHOD

******************************************************/
// this implements a greedy algorithm to find block boundaries
void GreedyFindByMDL(MDL_struct *B)
{
	int i,j, min_i, min_j, k,g;
	int Improvement=1;
	int *Bounds;  // just make a big array to use here
	int *TestBounds;
	int CNB;  // current number of boundaries
	int TNB;  // number of boundaries in "test" array
	double min, temp;
	int *ReValue;
	double Q, q = 75.0 / (double)B->NumMarkers;  // these and r are merely for
	int r;			// printing dots out when they ought to be

	
	
	Bounds = (int *)ECA_CALLOC((size_t)(B->NumMarkers + 3), sizeof(int));
	TestBounds = (int *)ECA_CALLOC((size_t)(B->NumMarkers + 3), sizeof(int));
	
	
	// Initialize Bounds to have the two endpoints:
	Bounds[0] = 0;
	Bounds[1] = B->NumMarkers;
	CNB = 2;
	
	
	// check to see if space has been allocated to the matrices.  If not, do so.
	if(B->E == NULL)
		B->E = AllocEntMat(B->NumMarkers);
	if(B->CanonSeqMultParCost == NULL)
		B->CanonSeqMultParCost = AllocEntMat(B->NumMarkers);
	if(B->SumOfCodeLengths == NULL)
		B->SumOfCodeLengths = AllocEntMat(B->NumMarkers);
	if(B->NT == NULL)
		B->NT = AllocNTMat(B->NumMarkers);
	
		
	// initialize both E and NT to -1's
	InitEntMat(B->E, B->NumMarkers);
	InitEntMat(B->CanonSeqMultParCost, B->NumMarkers);
	InitEntMat(B->SumOfCodeLengths, B->NumMarkers);
	InitNTMat(B->NT, B->NumMarkers);
	
	// and compute the code lengths from the filled data:
	ComputeCodeLengths(B, B->Y);
	


	
	
	
	
	while(Improvement) {  // enter a loop that you will keep doing so long as Improvement gets made
	
		
		// initialize variables that will tell us if we can do better,
		// and if so, which position is best:
		Improvement = 0;
		min = ModelBitsAndDataBits(CNB,Bounds,B);  
		min_i = -1;
		min_j = -1;
		TNB = CNB + 1;
		
		/*********   PRINTING OUT THE PROGRESS ************/
		if(gVerboseFlag)  {
			printf("\n\nCurrently %d Bounds at Score %.3f.  Testing More Positions\n  ",
					 CNB, min );
			r = 0;
			Q = 0.0;
			
//scanf(" %c", &boing);
		}
		/**************************************************/
		
		
		// now, we try out the different possible TestBounds arrays:
		for(i=0;i<CNB-1;i++)  {  // cycle over the different "spaces between"
							   // the current bounds.
			
			// get TestBounds set up:
			for(k=0;k<CNB;k++)  {
				TestBounds[k+(k>i)] = Bounds[k];
			}
			
			for(j=Bounds[i]+1;j<Bounds[i+1];j++)  {  // cycle over all positions within
													  // that "space between"  
				
				TestBounds[i+1] = j;  // make TestBounds have a new position in the
									  // appropriate "space between"
				
				// now, see how this one scores:
				temp =  ModelBitsAndDataBits(TNB,TestBounds,B);
				
				if(gPrintTestBounds) {
					fprintf(gPrintTestBoundsFile,"\nScore = %.4f,  (%d) Bounds = [",temp,TNB);
					for(g=0;g<TNB;g++) {
						fprintf(gPrintTestBoundsFile,"%d",TestBounds[g]);
						if(g<TNB-1)
							fprintf(gPrintTestBoundsFile,", ");
					}
					fprintf(gPrintTestBoundsFile,"]");
				}
				
				// and compare it to the current min
				if(temp < min  && (temp > 0.0))  {
					Improvement = 1;
					min_i = i;
					min_j = j;
					min = temp;
				}
				/*********   PRINTING OUT THE PROGRESS ************/
				if(gVerboseFlag)  {  
					Q += q;
					if((int)Q != r) {
						printf(".");
						fflush(stdout);
						r = (int)Q;
					}
				}
				/**************************************************/
			}
			
		}  // closes the i-loop over the current bounds

		
		// now that we are down here, min_i and min_j should tell us where the best
		// place to put a new boundary was.  So, if we made improvement, then we shall
		// set Bounds to this new array of positions...
		if(Improvement)  {
			// get TestBounds set up as it was when the min was found
			for(k=0;k<CNB;k++)  {
				TestBounds[k+(k>min_i)] = Bounds[k];
			}
			TestBounds[min_i+1] = min_j;
			
			// then increment CNB, and copy that over to Bounds
			CNB++;
			for(k=0;k<CNB;k++)  {
				Bounds[k] = TestBounds[k];
			}
		}
	
	}  
	
	// once out here, we return the array of boundaries, and also the number of
	// boundaries in there.  
	B->NumBounds = CNB;
	ReValue = (int *)ECA_CALLOC((size_t)CNB,sizeof(int));
	for(i=0;i<CNB;i++)  {
		ReValue[i] = Bounds[i];
	}
	
	
	B->Bounds = ReValue;
	
	// clean up the memory allocated.
	free(Bounds);
	free(TestBounds);
		
	
}


// this computes the sum of the model bits and the data bits, given n boundaries 
// (including first and last) given in A.
// To do this, the function uses Entropies and number of types, which might
// be stored in B->E and B->NT.  If either of those is < 0 for  a particular
// block, this function calls another, that will compute the entropy and the number
// of types for that block, and it stores those values, in case they need to be used
// again.  
double ModelBitsAndDataBits(int n, int *A, MDL_struct *B) 
{
	int i, a, b;  // a is the column specifier and b is the row specifier
	double value;
	
	// model parameters bit:
	value = log(B->NumMarkers)/log(2.0);  // The bits for the number of markers
	for(i=0;i<n-1;i++)  {  // cycle only to < n-1 because 0 is given as a boundary already.
		value += log(B->NumMarkers-i)/log(2.0);  // the bits for the positions of the markers
	}
	
/*	
	// this here codes the bits for the positions of the markers as differences.  This is also
	// where the length of each sequence is sent (that is not assumed known).
	for(i=1;i<n;i++)  {
		value += log(A[i] - A[i-1])/log(2.0);  // the bits for the positions of the markers
	}
*/
	
	
	// cycle over the blocks
	for(i=0;i<n-1;i++)  {
		
		a = A[i];
		b = A[i+1] - 1;
		
		if( B->E[b][a] < 0  || gCloserLook)  {  // if the necessary quantities have not
												  // been computed, then compute them and store them
			printf("\n\nBummer Dude!  Lacking an entry in Entropy or NT Matrix\n\n");
		}
		
		// add to it the part for specifying the blocks
//		value += (double)(B->NT[b][a] * (b - a + 1));
//		value += (double)B->NT[b][a] * log(B->NT[b][a] + 1) / log(2.0) + (b-a+1) * B->CanonSeqMultParCost[b][a];
		value += B->SumOfCodeLengths[b][a];
		
		// then add the part for sending the data
		value += (double)B->NumSeq * B->E[b][a];
		
	}
	
	return(value);
	
}




/*********************************

COMPUTING CODE LENGTHS

*********************************/
// this computes the code lengths in bits for 0's and 1's in the
// data at each site.  It just counts up all the 0's and 1's and
// computes the frequency from those.  This can be applied to data
// that does or does not have missing data.  Depending on if you 
// pass Y or Yobs to it. 
void ComputeCodeLengths(MDL_struct *B, int **Data)
{
	int i,j;
	double Total;
	
	if(B->CodeLengths == NULL)  {  // allocate memory to this if necessary
		B->CodeLengths = (double **)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(double *));
		for(i=0;i<B->NumMarkers;i++)  {
			B->CodeLengths[i] = (double *)ECA_CALLOC( (size_t)2, sizeof(double));
		}		
	}
	
	// now cycle over the sites and take care of the counting
	for(j=0;j<B->NumMarkers;j++)  {
		B->CodeLengths[j][0] = 0.0;
		B->CodeLengths[j][1] = 0.0;
		for(i=0;i<B->NumSeq;i++)  {
			if(Data[i][j]==0)  {
				B->CodeLengths[j][0] += 1.0;
			}
			else if(Data[i][j]==1)  {
				B->CodeLengths[j][1] += 1.0;
			}
		}
		Total =  B->CodeLengths[j][0] + B->CodeLengths[j][1];
		if(B->CodeLengths[j][0] != Total && B->CodeLengths[j][0] != 0) {
			// now, the code length calculation
			B->CodeLengths[j][0] = -log(B->CodeLengths[j][0]/Total) / log(2.0);
			B->CodeLengths[j][1] = -log(B->CodeLengths[j][1]/Total) / log(2.0);
		}
		else {
			B->CodeLengths[j][0] = 0.0;
			B->CodeLengths[j][1] = 0.0;
		}
		
/*		printf("\n%d:  %f  %f", j, B->CodeLengths[j][0], B->CodeLengths[j][1]); */
	}
//scanf("%d",&i);
}


/***********************************
	SORTING THE BLOCK TYPES, ETC
***********************************/
// compare tree tips (in our tree for counting haplotypes)  on the basis of 
// which one has more visits through it
int CompareBinNodesForVisitSort(const void *vT1, const void *vT2)
{
	int value = 0;
	
	const bin_node **T1p = (const bin_node **)vT1;
	const bin_node **T2p = (const bin_node **)vT2;
	
	const bin_node *T1 = *(T1p);
	const bin_node *T2 = *(T2p);
	
	if(T1->Visits < T2->Visits) 
		value = 1;
	if(T1->Visits > T2->Visits)
		value = -1;
		
	return(value);
}




/******************************************
FOR IMPLEMENTING THE DP ALGORITHM
******************************************/
// this fills in all the matrices that need to be filled
// in to implement the DP algorithm.  
/* void FillInMatrices(MDL_struct *B)
{
	int a,b;
	
	// check to see if space has been allocated to the matrices.  If not, do so.
	if(B->E == NULL)
		B->E = AllocEntMat(B->NumMarkers);
	if(B->CanonSeqMultParCost == NULL)
		B->CanonSeqMultParCost = AllocEntMat(B->NumMarkers);
	if(B->SumOfCodeLengths == NULL)
		B->SumOfCodeLengths = AllocEntMat(B->NumMarkers);
	if(B->NT == NULL)
		B->NT = AllocNTMat(B->NumMarkers);
	
		
	// initialize both E and NT to -1's
	InitEntMat(B->E, B->NumMarkers);
	InitEntMat(B->CanonSeqMultParCost, B->NumMarkers);
	InitEntMat(B->SumOfCodeLengths, B->NumMarkers);
	InitNTMat(B->NT, B->NumMarkers);
	
	// and compute the code lengths from the filled data:
	ComputeCodeLengths(B, B->Y);
	
	// then cycle over the starting points and ending points of blocks
	for(a=0;a<B->NumMarkers;a++)  {
		if(gVerboseFlag)
			printf("\nComputing Matrices, a = %d",a);
		for(b=a;b<B->NumMarkers;b++)  {
			ComputeBlockEntropyAndNT_WithSeqs(a,b,B);
		}
	}
	
	
	

}
*/






/******************************************
For resolving cycles in the iteration over gR
******************************************/

void ResolveCycle(int *Rvisited, double *Rscores, MDL_struct *B){

  int R1,R2,minR;
  double minScore;

  printf("\nThe iteration over assumed values of R is not converging.\n");
  printf("The iteration is cycling over the assumed values: ");

  R1=gR;
  R2=-1;
  minR=R1;
  minScore=Rscores[R1];
  while(R2!=gR){
    R2=Rvisited[R1];
    printf("%d (w/ score %f) ",R2,Rscores[R2]);    
    if(Rscores[R2]<minScore){
      minR=R2;
    }
    R1=R2;
  }
  
  printf("\nOf these values, assuming R=%d leads to the lowest description length.\n",minR);
  gR=minR;

  // Step through the cycle and find the value of gR in the cycle with the lowest bitscore
  // Set gR to this value
  if(gGreedy)
    GreedyBlocks(B);
  else
    DP_Blocks(B);


}









