/*
Copyright (c). The Regents of the
University of California (Regents). All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby granted,
provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
for commercial licensing opportunities. Created by Eric C. Anderson and
John A. Novembre,
Department of Integrative Biology, University of California, Berkeley.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ranlib.h"

#include "ECA_MemAlloc.h"
#include "MathStatRand.h"

#include "MDL_Blocks.h"


/*********************************************

DEALING WITH MISSING DATA

*********************************************/
// a function to look at the rates of missingness at each SNP separately
int *NumMissingByColumn(MDL_struct *B, int TypeOfMissing)
{
	int i,j, *temp = (int *)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(int));
	
	for(j=0;j<B->NumMarkers;j++)  {
		temp[j] = 0;
		for(i=0;i<B->NumSeq;i++)  {
			if(B->Yobs[i][j] == TypeOfMissing)
				temp[j]++;
		}
	}
	return(temp);
}



// this function collates the missing data of type TypeOfMissing (i.e. Ambiguous heterozygosity
// versus truly missing...) into a missing data info struct.
missing_data_info *CollectMissingSites(MDL_struct *B, int TypeOfMissing)
{
	
	int i,j;
	int NSWM;  // number of sequences with missing data
	int temp;
	missing_data_info *MDI = (missing_data_info *)ECA_MALLOC(sizeof(missing_data_info));
	
	
	// first off, initialize Y to the values in Yobs
	for(i=0;i<B->NumSeq;i++)  {
		for(j=0;j<B->NumMarkers;j++)  {
			B->Y[i][j] = B->Yobs[i][j];
		}
	}
	
	// then make the lists of the missing holes.
	// first, count NSWM:
	NSWM = 0;
	for(i=0;i<B->NumSeq;i++)  {
		for(j=0;j<B->NumMarkers;j++) {
			if(B->Y[i][j] == TypeOfMissing) {
				NSWM++;
				break;
			}
		}
	}
	
	// only do the rest if NSWM > 0
	if(NSWM > 0)  {
		
		// then allocate to the lists
		MDI->NumMissingArray = (int *)ECA_CALLOC((size_t)NSWM, sizeof(int));
		MDI->M_Idx = (int *)ECA_CALLOC((size_t)NSWM, sizeof(int));
		MDI->MissingIndices = (int **)ECA_CALLOC((size_t)NSWM, sizeof(int*));
		
		// cycle over things again and fill the M_Idx array
		NSWM = 0;
		for(i=0;i<B->NumSeq;i++)  {
			for(j=0;j<B->NumMarkers;j++) {
				if(B->Y[i][j] == TypeOfMissing) {
					MDI->M_Idx[NSWM++] = i;
					break;
				}
			}
		}
		
		
		// then cycle over the missing ones, and count how many missing sites there are
		MDI->TotalMissing = 0;
		for(i=0;i<NSWM;i++)  {
			temp = 0;
			for(j=0;j<B->NumMarkers;j++)  {
				if(B->Y[MDI->M_Idx[i]][j] == TypeOfMissing)  {
					temp++;
				}
					
			}
			MDI->NumMissingArray[i] = temp;
			MDI->TotalMissing += temp;
		}
		
		
		// then allocate to the MissingIndices elements and fill those in
		for(i=0;i<NSWM;i++)  {
			MDI->MissingIndices[i] = (int *)ECA_CALLOC((size_t)MDI->NumMissingArray[i],sizeof(int));
			temp = 0;
			for(j=0;j<B->NumMarkers;j++)  {
				if(B->Y[MDI->M_Idx[i]][j] == TypeOfMissing)  {
					MDI->MissingIndices[i][temp++] = j;
				}
			}
		}
		
	}
	else {
		MDI->TotalMissing = 0;
	}
	
	MDI->NSWM = NSWM;
	
	// now return the thing
	return(MDI);
	
}


// fill an array of the neighbor indices.  Assumes the array has already been
// allocated to.  Every NN consecutive elements of the array correspond to one
// position of a sliding window on the array.  The elements in the array are the 
// indexes of the markers included in that sliding window position.  
// The sliding window positions may be indexed by the starting position s between 0
// and NN inclusive.  
// If a -1 appears
// in that sliding window position, then the window has slid off the edge from that
// position.  
// I is the index of the chromosome.  M is the index of the marker for whom we wish
// to designate the neighbors.
// n is the length of the neighbors array (for filling it in with -1's
void DesignateNeighbors(int I, int M, int NN, int *Neighbors, int n, MDL_struct *B)
{
	int NonMis;   // to count up the number of non-missing sites
	int i,j;
	
	
	//  fill Neighbors with -1's to start
	for(i=0;i<n;i++)  {
		Neighbors[i] = -1;
	}
	
	// then work back to the NN-th left neighbor, or to 0 which ever comes first
	NonMis=0;
	j = M;
	while(NonMis < NN && j > 0) {
		j--;  // decrement j
		if(B->Y[I][j] >= 0)  {  // if it ain't missing
			// add this marker index to Neighbors in the right spot
			NonMis++;
			Neighbors[NN-NonMis] = j;
		}
	}
	
	// then work forward to the NN-th right neighbor or to 0, which ever comes first
	NonMis = 0;
	j = M;
	while(NonMis < NN && j < B->NumMarkers-1) {
		j++;  // increment j
		if(B->Y[I][j] >= 0)  {  // if it ain't missing
			// add this marker index to Neighbors in the right spot
			NonMis++;
			Neighbors[NN+NonMis-1] = j;
		}
	}
}



// for chromosome I, missing at marker M, and with non-missing neighbor sites 
// designated in Neighbors, compute the probability the the Marker at M in I is
// a 0 or a 1 for the NN different positions of the sliding window, and given 
// OneInit and ZeroInit as pseudo-counts to initialize things with.  Return the
// output in OneProbs and ZeroProbs which are arrays indexed by the sliding window
// starting positions:  0,...,NN, inclusive.
// This compares things to what exists in Y rather than Yobs.   
void ComputeProbsForHoleY(int I, int M, int NN, int *Neighbors, MDL_struct *B, 
						double ZeroInit, double OneInit, double *ZeroProbs, double *OneProbs)
{
	int i,k,n;
	int NoNegs;
	int NonMissers[MAX_NEIGHBORS * 2];  // temporary to store the values at the neighbors
	int Match, Miss, Bomp;  // counts the number of sites in a comparison sequence that Match,
							// that are Missing, and which do not match.  (Bomp!)
	double TempOne, TempZero, normo;
	
	for(k=0;k<=NN;k++)  {  // cycle over the different window starting positions.
		ZeroProbs[k] = ZeroInit;
		OneProbs[k] = OneInit;
		
		// for each window starting position, count up the matching or consistent sequences,
		// only if there are no -1's amongst the Neighbors.
		
		// transfer the non-missers to NonMissers.  I think that will be faster
		// than checking the original individual for each comparison
		for(n=0;n<NN*2;n++)  {
			if(Neighbors[n] >= 0)
				NonMissers[n] = B->Y[I][Neighbors[n]];
			else
				NonMissers[n] = -100;  // these values should never be used for anything anyway
		}
		
		// check for -1's
		NoNegs = 1;
		for(n=k;n<k+NN;n++) {
			if(Neighbors[n] < 0)  {
				NoNegs = 0;
				break;
			}
		}
		if(NoNegs)  {  // here, only doing it if there were no -1's in the neighbors
			for(i=0;i<B->NumSeq;i++)  {
				if(B->Y[i][M] >= 0)  {  // only count those that don't have missing data 
										// at the same site.
										
					// initialize to count
					Match = 0;  Miss = 0;  Bomp = 0;
					for(n=k;n<k+NN;n++) {  // cycle over the neighbors in the k-th window position
						if(B->Y[i][Neighbors[n]] == NonMissers[n])
							Match++;
						else if(B->Y[i][Neighbors[n]] < 0)
							Miss++;
						else {
							Bomp++;
							break;
						}
					}
					
					if(Bomp==0)  {  // only do this if things match or miss everywhere.
						if(B->Y[i][M] == 0)  {
							TempZero = (double)Match / (double)(Miss + Match);
							TempOne = 0.0;
						}
						else if(B->Y[i][M] == 1)  {
							TempOne = (double)Match / (double)(Miss + Match);
							TempZero = 0.0;
						}
						else {
							TempOne = 0.0;
							TempZero = 0.0;
						}
						// now, add those into the Probs Arrays
						ZeroProbs[k] += TempZero;
						OneProbs[k] += TempOne;
						
					}
					
				}  // closes the conditional that the sequences is non-missing at M
				
			}  // closes the i-loop over the other sequences
		
		
		}  // closes the conditional if(NoNegs)
		
		// now, normalize the entries in the Probs arrays
		normo = ZeroProbs[k] + OneProbs[k];
		ZeroProbs[k] /= normo; 
		OneProbs[k] /= normo;
		
		
	}  // closes the k-loop over window positions
	
	// that should do it.
} 


// given arrays of probabilities for zero or one from NN+1 different
// window positions, indexed by 0,...,NN, inclusive, 
// for 0's or 1's from an ambig het missing site, werandomly choose
// and return a zero if it is 01 and a 1 if it is 10.  This is for
// AmbigHet data.  Assumes that  
int DrawFromMostCertainAmbigH(double *ZP1, double *OP1, double *ZP2, 
		double *OP2, int NN)
{
	int k;
	int Max = 0;
	double temp, MaxD = 0.0, rando;
	int TheValue;
	double temp01, temp10, normo;
	double MaxTemp01, MaxTemp10;
	
	// first, determine which of the window positions is most certain.  This is
	// the k for which fabs(ZeroProbs[k] - OneProbs[k]) is greatest.
	for(k=0;k<=NN;k++)  {	
		temp01 = ZP1[k] * OP2[k];
		temp10 = ZP2[k] * OP1[k];
		normo = temp01 + temp10;
		temp01 /= normo;
		temp10 /= normo; 
		
		temp = fabs(temp01 - temp10);
		if(temp > MaxD)  {
			Max = k;
			MaxD = temp;
			MaxTemp01 = temp01;
			MaxTemp10 = temp10;
		}
	}
	
	// then choose the value
	rando = (double)ranf();
	
	if(rando < MaxTemp01)  {
		TheValue = 0;
	}
	else  {
		TheValue = 1;
	}
	
	return(TheValue);
	
}



// fills a hole caused by ambiguous heterozygosity using the sliding window
// method.  THIS FUNCTION ASSUMES THAT EACH PAIR INVOLVED IN THE AMBIG
// HET HAS SEQUENCES WITH THE indexes given by V and V+1 where V%2==0.   
void FillAmbigHetHole(MDL_struct *B, int V, int J, int NN)
{
	int M, I, I2;  // to store the index of the missing site, and of the indiv respectively
	int Neighbors1[2 * MAX_NEIGHBORS];  //  Neighbors array. This holds the indexes of the markers
									   // that get used as neighbors
	int Neighbors2[2 * MAX_NEIGHBORS];
	int Result;
									   
	double ZP1[MAX_NEIGHBORS], OP1[MAX_NEIGHBORS];  // for zero and one probs
	double ZP2[MAX_NEIGHBORS], OP2[MAX_NEIGHBORS];  // for zero and one probs
	
	// do some error checking:
	if(V%2 != 0)  {
		printf("\nWarning!!  V mod 2 not zero in FillAmbigHetHole()\n\n");
	}
	
	
	// first, fill the Neighbors array.  The elements from 0,...,NN-1 should
	// be the NN non-missing neighbors to the left of the point in question, and then
	// NN to 2NN-1 will be the right neighbors.
	
	// So, first get the index of the missing site
	I = B->AH->M_Idx[V];
	I2 =  B->AH->M_Idx[V+1];
	
	if(I%2 != 0)  {
		printf("\nWarning!!  I not zero in FillAmbigHetHole()\n\n");
	}
	if(I2 != I+1)  {
		printf("\nWarning!!  I2 not I+1 in FillAmbigHetHole()\n\n");
	}
	
	M = B->AH->MissingIndices[V][J];
	
	DesignateNeighbors( I,  M,  NN, Neighbors1, 2 * MAX_NEIGHBORS, B);
	
	DesignateNeighbors( I2,  M,  NN, Neighbors2, 2 * MAX_NEIGHBORS, B);
	
	ComputeProbsForHoleY( I,  M,  NN,  Neighbors1, B, 
						.000001, .000001,
						 ZP1, OP1);
						 
	ComputeProbsForHoleY( I2,  M,  NN,  Neighbors2, B, 
						.5, .5,
						 ZP2, OP2);
						 
	Result = DrawFromMostCertainAmbigH(ZP1, OP1, ZP2, OP2, NN);
	
	if(Result==0)  {
		B->Y[I][M] = 0;
		B->Y[I2][M] = 1;
	}
	else {
		B->Y[I][M] = 1;
		B->Y[I2][M] = 0;
	}
}



// Fills in a random hole caused by unresolved heterosygosity.  
// Returns the number of 
// holey sequences still remaining.  This way you can run
// it within a while loop that exits once it has filled all the holes.  
// THIS FUNCTION TRIMS THE ARRAYS DOWN and frees some memory each
// time it runs
int FillRandomAmbigHet(MDL_struct *B, int NN)
{
	int V,J;  // for the individual and the marker indices in TM
	
	if(B->AH->NSWM == 0)
		return(0);
		
		
	// choose a V and then J randomly (but not uniformly here!---currently this gives
	// higher probability to those holes in sequences that have fewer missing holes.
	V = UniformRV(0,B->AH->NSWM-1);
	J = UniformRV(0,B->AH->NumMissingArray[V]-1);
	
	// if V%2 == 1, then decrease V by one
	if(V%2 == 1)  {
		V--;
	}
	
	FillAmbigHetHole(B, V, J, NN);
	
	// now we have to deal with deleting all the elements from the arrays that
	// we need to alter:
	
	// first an element from the array of Missing Indices.
	DeleteIntElement(B->AH->MissingIndices[V], J, B->AH->NumMissingArray[V]);
	B->AH->NumMissingArray[V] -= 1;
	
	DeleteIntElement(B->AH->MissingIndices[V+1], J, B->AH->NumMissingArray[V+1]);
	B->AH->NumMissingArray[V+1] -= 1;
	
	
	
	// then, if that was the last remaining ambig het site in indiv V, we have
	// to delete that individual from NumMissingArray and also from MissingIndices, and
	// from M_Idx, but we only want to decrement NSWM once
	if(B->AH->NumMissingArray[V] == 0  || B->AH->NumMissingArray[V+1] == 0 ) {
		if(B->AH->NumMissingArray[V] == 0 && B->AH->NumMissingArray[V+1] == 0 )  {  // make sure it is missing at both
			
			// delete the stuff for the first sequence of the pair
			DeleteIntElement(B->AH->NumMissingArray, V,B->AH->NSWM);
			DeleteIntElement(B->AH->M_Idx, V,B->AH->NSWM);
			FreeAndDeleteIntStarElement(B->AH->MissingIndices, V, B->AH->NSWM);
			// decrement the number of sequences with missing data:
			B->AH->NSWM--;
			
			// then do the same with the second member of the pair:
			DeleteIntElement(B->AH->NumMissingArray, V,B->AH->NSWM);
			DeleteIntElement(B->AH->M_Idx, V,B->AH->NSWM);
			FreeAndDeleteIntStarElement(B->AH->MissingIndices, V, B->AH->NSWM);
			B->AH->NSWM--;
			
		}
		else {
			printf("\n\nTotally Bummin' Dude!! Different numbers of Ambig Het sites\n");
			printf("in FillRandomAmbigHet()\n\n");
		}
	}
	
	return(B->AH->NSWM);
}



// given array of probabilities for zero or one from NN+1 different
// window positions, indexed by 0,...,NN, inclusive, randomly choose
// and return a zero or a one.  This is for "Truly Missing Data" 
int DrawFromMostCertainSingle(double *ZeroProbs, double *OneProbs, int NN)
{
	int k;
	int Max = 0;
	double temp, MaxD = 0.0, rando;
	int TheValue;
	
	// first, determine which of the window positions is most certain.  This is
	// the k for which fabs(ZeroProbs[k] - OneProbs[k]) is greatest.
	for(k=0;k<=NN;k++)  {
		temp = fabs(ZeroProbs[k] - OneProbs[k]);
		if(temp > MaxD)  {
			Max = k;
			MaxD = temp;
		}
	}
	
	// then choose the value
	rando = (double)ranf();
	
	if(rando < ZeroProbs[Max])  {
		TheValue = 0;
	}
	else  {
		TheValue = 1;
	}
	
	return(TheValue);
	
}



// fill the J-th missing marker in the V-th individual that has something missing in it.
// Fill the hole based on the NN neighbors that are not missing.
void FillTrulyMissingHole(MDL_struct *B, int V, int J, int NN)
{
	int M, I;  // to store the index of the missing site, and of the indiv respectively
	int Neighbors[2 * MAX_NEIGHBORS];  //  Neighbors array. This holds the indexes of the markers
									   // that get used as neighbors
									   
	double ZP[MAX_NEIGHBORS], OP[MAX_NEIGHBORS];  // for zero and one probs
	
	// first, fill the Neighbors array.  The elements from 0,...,NN-1 should
	// be the NN non-missing neighbors to the left of the point in question, and then
	// NN to 2NN-1 will be the right neighbors.
	
	// So, first get the index of the missing site
	I = B->TM->M_Idx[V];
	M = B->TM->MissingIndices[V][J];
	
	DesignateNeighbors( I,  M,  NN, Neighbors, 2 * MAX_NEIGHBORS, B);
	
	ComputeProbsForHoleY( I,  M,  NN,  Neighbors, B, 
						.000001, .000001,
						 ZP, OP);
						 
						 
	B->Y[I][M] = DrawFromMostCertainSingle(ZP, OP, NN);
	
}




// in a random order this will fill in all the holes that
// are due to missing genotypes.  Returns the number of 
// holey sequences still remaining.  This way you can run
// it within a while loop that exits once it has filled all the holes.  
int FillRandomTrulyMissing(MDL_struct *B, int NN)
{
	int V,J;  // for the individual and the marker indices in TM
	
	if(B->TM->NSWM == 0)
		return(0);
		
		
	// choose a V and then J randomly (but not uniformly here!---currently this gives
	// higher probability to those holes in sequences that have fewer missing holes.
	V = UniformRV(0,B->TM->NSWM-1);
	J = UniformRV(0,B->TM->NumMissingArray[V]-1);
	
	FillTrulyMissingHole(B, V, J, NN);
	
	// now we have to deal with deleting all the elements from the arrays that
	// we need to alter:
	
	// first an element from the array of Missing Indices
	DeleteIntElement(B->TM->MissingIndices[V], J, B->TM->NumMissingArray[V]);
	B->TM->NumMissingArray[V] -= 1;
	
	// then, if that was the last remaining truly missing site in indiv V, we have
	// to delete that individual from NumMissingArray and also from MissingIndices, and
	// from M_Idx, but we only want to decrement NSWM once
	if(B->TM->NumMissingArray[V] == 0) {
		DeleteIntElement(B->TM->NumMissingArray, V,B->TM->NSWM);
		DeleteIntElement(B->TM->M_Idx, V,B->TM->NSWM);
		FreeAndDeleteIntStarElement(B->TM->MissingIndices, V, B->TM->NSWM);
		
		// then decrement the number of sequences with missing data:
		B->TM->NSWM--;
	}
	
	return(B->TM->NSWM);
}



// Delete the x-th element from an array of ints with n elements,
// and shift everything to the left.  Does NOT decrement n
void DeleteIntElement(int *A, int x, int n)
{
	int i;
	
	if(x < 0 || x >= n)  // do nothing if we are asking to delete something
						  // that is not there
		return;
	
	// decrement n by one.  Does not change the real n (remember---call by value)
	n -= 1;
	
	// shift everything
	for(i=x;i<n;i++)  {
		A[i] = A[i+1];
	}
	
	// that should be it
}


// Delete the x-th element from an array of int *'s with n elements,
// and shift everything to the left.  Does NOT decrement n
void FreeAndDeleteIntStarElement(int **A, int x, int n)
{
	int i;
	
	if(x < 0 || x >= n)  // do nothing if we are asking to delete something
						  // that is not there
		return;
		
	// free the memory allocated to the x-th element
	if(A[x] != NULL)
		free(A[x]);
	
	// decrement n by one.  Does not change the real n (remember---call by value)
	n -= 1;
	
	// shift everything
	for(i=x;i<n;i++)  {
		A[i] = A[i+1];
	}
	
	// that should be it
}


void FillAllMissingHoles(MDL_struct *B)
{
	int DotInterval, i;
	
	
	// first collect all the information about missing data
	B->AH = CollectMissingSites(B, AMBIG_HET);
	B->TM = CollectMissingSites(B, MISSING_GENOTYPE);
		
	
	// first fill all the ambiguous het ones
	if(gVerboseFlag)
		printf("\n\nFilling %d Unresolved Heterozygosity Holes:\n",B->AH->TotalMissing);
	DotInterval = B->AH->TotalMissing / 150;
	i=0;
	while(FillRandomAmbigHet(B, B->AH_WindowWidth)) {
		i++;
		if(gVerboseFlag)
			if(i % DotInterval == 0) {
				printf(".");
				fflush(stdout);
			}
	}
	// then deallocate the B->AH struct:
	DeAllocMissingDataInfo(B->AH);
	
	
	// then fill all the truly missing ones
	if(gVerboseFlag)
		printf("\n\nFilling %d Missing Data Holes:\n",B->TM->TotalMissing);
	DotInterval = B->TM->TotalMissing / 75;
	i=0;
	while(FillRandomTrulyMissing(B, B->TM_WindowWidth)) {
		i++;
		if(gVerboseFlag)
			if(i % DotInterval == 0) {
				printf(".");
				fflush(stdout);
			}
	}
	// then deallocate the B->AH struct:
	DeAllocMissingDataInfo(B->TM);

}
