/*
Copyright (c). The Regents of the
University of California (Regents). All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research, and not-for-profit purposes,
without fee and without a signed licensing agreement, is hereby granted,
provided that the above copyright notice, this paragraph and the
following two paragraphs appear in all copies, modifications, and
distributions. Contact The Office of Technology Licensing, UC Berkeley,
2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
for commercial licensing opportunities. Created by Eric C. Anderson and
John A. Novembre,
Department of Integrative Biology, University of California, Berkeley.

IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY,
PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION TO
PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
*/




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>


#include "ECA_MemAlloc.h"
#include "MathStatRand.h"

#include "MDL_Blocks.h"


/* a little global variable with scope within this file */
double **gStirlingArray = NULL;

// to sort into descending order
int CompIndexedDoubles(const void *A, const void *B)
{
	indexed_doubles_struct *DA = (indexed_doubles_struct *)A, 
		*DB = (indexed_doubles_struct *)B;
		
	if(DB->d == DA->d)
		return(0);
	else if(DA->d > DB->d)
		return(-1);
	else
		return(1);
	
}

// to sort into descending order
int CompIndexedDoublesAscendng(const void *A, const void *B)
{
	indexed_doubles_struct *DA = (indexed_doubles_struct *)A, 
		*DB = (indexed_doubles_struct *)B;
		
	if(DB->d == DA->d)
		return(0);
	else if(DA->d > DB->d)
		return(1);
	else
		return(-1);
	
}

/* return the data and model costs in [a,b] given that the previous
block is [z,a-1]   */ 
double CostWithMixture(int z, int a, int b, MDL_struct *B, double *DataCostOutput, double *DeltaCost, 
			double *PCost, int *NumDelts, double *log2NabSum, double *FixedBitCost, double *KLD)
{
	int i,j,r,*Zseq, *Aseq;  // canonical group assignments of the seqs in [z,a-1] and [a,b] 
	double *Freqs, *MLogFreqs;  // the frequencies of the different types and neg logs of those freqs
								// in the [a,b] block
	indexed_doubles_struct *D;
	trans_info_struct *tz = &(B->Tr[a-1][z]);
	trans_info_struct *ta = &(B->Tr[b][a]);
	
	double DataIndCost;  // will store the cost per row of sending the data via the Independence model
	int NumSpec, // the number of cells specified in the row transition matrix
		*Spec;  // array to keep track of which cells have been specified.  0 means no, 1 means yes
	int *NumInTA;  // the number of sequences of this row that fall into the canon seq groups of [a,b]
	
	// variables for computing the cost of the model under dependence
	double  NumAccountedFor, tempNumAccountedFor,
			PosSpecSum, NumSpecSum, tempPosSpecSum, tempNumSpecSum, ModelCost, tempModelCost,
			SpecCodeLength, tempSpecCodeLength, prop, UnSpecCodeLength, 
			DataCost, tempDataCost;
	double log2Nab;
			
	double TotalCostOverAllRows = 0.0;
	double *pstar, KLweight, KLD_r;  
	
	
	// allocate memory as needed
	Freqs = (double *)ECA_CALLOC((size_t)ta->N, sizeof(double));
	MLogFreqs = (double *)ECA_CALLOC((size_t)ta->N, sizeof(double));
	Zseq = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
	Aseq = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
	D = (indexed_doubles_struct *)ECA_CALLOC( (size_t)ta->N, sizeof(indexed_doubles_struct));
	Spec = (int *)ECA_CALLOC( (size_t)ta->N, sizeof(int));
	NumInTA = (int *)ECA_CALLOC( (size_t)ta->N, sizeof(int));
	pstar = (double *)ECA_CALLOC( (size_t)ta->N,sizeof(double));
	
	// then fill the seqs indices
	for(i=0;i<tz->N;i++)  {
		for(j=0;j<tz->Nd[i];j++)  {
			Zseq[tz->Idx[i][j]] = i;
		}
	}
	for(i=0;i<ta->N;i++)  {
		for(j=0;j<ta->Nd[i];j++)  {
			Aseq[ta->Idx[i][j]] = i;
		}
	}
	
	// then fill the Freqs and MLogFreqs arrays
	for(i=0;i<ta->N;i++)  {
		Freqs[i] = (double)ta->Nd[i]/(double)B->NumSeq;
		MLogFreqs[i] = -log(Freqs[i])/log(2.0);
	}
	
	
	/* initialize the output parameters to collect a sum */
	*DataCostOutput = 0.0;
	*DeltaCost = 0.0;
	*PCost = 0.0;
	*NumDelts = 0;
	*log2NabSum=0.0;
	*FixedBitCost=0.0;  /* for the bits that say whether NumSpec==0 or is >0 */
	*KLD = 0.0;  /* Kullback-Leibler divergence between P* rows and q */
	
	log2Nab = log((double)ta->N) / log(2.0);
	
	// then cycle over the rows using r
	for(r=0;r<tz->N;r++)  {
	
		// initialize the indexed doubles struct
		for(i=0;i<ta->N;i++)  {
			D[i].Idx = i;
			D[i].d = 0.0;
		}
		
		// then fill them with the codelengths, summed within each canon seq class 
		// of the sequences in row r
		DataIndCost = 0.0;  for(j=0;j<ta->N;j++)  NumInTA[j] = 0;  // initialize to accumulate sums
		for(j=0;j<tz->Nd[r];j++)  {   /* NOTE: tz->Idx[r][j]  is the index of the j-th sequence (out of Nd[r] that is 
										of type r in the previous (z) block.  And Aseq and Zseq give the types in the A and
										the Z block respectively, of the sequences with a certian index.  It seems like this might be 
										redundant in some weird way, but oh well!  */
			
			
			D[ Aseq[tz->Idx[r][j] ] ].d += MLogFreqs[ Aseq[tz->Idx[r][j] ] ];   /* NOTE: in the end each D[j] will be proportional to the 
																					amount the j-th type in block k (i.e. the j-th element
																					in the row) contributes to the Kullback-Leibler divergence
																					between q and the observed proportion of transitions.   */
			DataIndCost += MLogFreqs[ Aseq[tz->Idx[r][j] ] ];
			NumInTA[Aseq[tz->Idx[r][j] ]]  += 1;
		}
		
		// then sort D
		qsort((void *)D, (size_t)ta->N, sizeof(indexed_doubles_struct), CompIndexedDoubles);
		
		// now we start trying out the cost of sending the data and the model for the dependence
		// version, working our way down from the costliest to the least costly cells (as they have
		// been ordered). 
		NumSpec = 0;  for(i=0;i<ta->N;i++)  Spec[i] = 0;  // initialize some variables
		NumAccountedFor = 0.0; tempNumAccountedFor = 0.0;
		PosSpecSum = 0.0; tempPosSpecSum = 0.0; NumSpecSum = 0.0; tempNumSpecSum = 0.0;
		SpecCodeLength = 0.0;  tempSpecCodeLength = 0.0;
		ModelCost = 0.0;  // assume that the cost for sending the multinomial model is zero
						  // because it has already been sent.
		DataCost = DataIndCost;  // the cost of sending the data is initialized to the cost 
							     // of sending the data given the independence model.
		
		// then actually compute
		for(j=0; j<ta->N-1 && D[j].d > 0.0; j++)  {
			tempNumAccountedFor = NumAccountedFor + (double)ta->Nd[D[j].Idx];
			
			// compute the overall model cost that would transpire if we specified the j-th cell
			tempNumSpecSum = NumSpecSum + log(ta->N - NumSpec)/log(2.0);
			tempPosSpecSum = PosSpecSum + log(B->NumSeq - tempNumAccountedFor)/log(2.0);
			tempModelCost = tempNumSpecSum + tempPosSpecSum;
			
			// now we have to compute the cost of sending the data.  This will depend
			// on the codelengths given the specified cells, and the codelengths given the
			// unspecified cells.
			
			// here is the code length for all the specified cells
			tempSpecCodeLength = SpecCodeLength + (double)NumInTA[D[j].Idx] * 
					(-1) * log( (double)NumInTA[D[j].Idx]/(double)tz->Nd[r] ) / log(2.0);
					
			// and here is the code length for all the cells that are unspecified.
			// you have to change their freqs a bit so that everything sums to one
			// even with the new frequencies that have been specified.
			// find the proportion to change them by:
			prop = 0.0;
			for(i=0;i<ta->N;i++)  {
				if(Spec[i]==0 && i != D[j].Idx)
					prop += Freqs[i];
			}
			prop /= ((double)B->NumSeq - tempNumAccountedFor)/(double)(B->NumSeq);
			prop = 1.0/prop;
			
			// then actually compute the code-lengths:
			UnSpecCodeLength = 0.0;
			for(i=0;i<ta->N;i++)  {
				if(Spec[i]==0 && i != D[j].Idx)  {
					UnSpecCodeLength -= (double)NumInTA[i] * log(prop * Freqs[i]) / log(2.0);
				}
			}
			tempDataCost = tempSpecCodeLength + UnSpecCodeLength;
			
			// now we are able to compare and see if there is a benefit to modeling the 
			// dependence  
			if(ModelCost + DataCost > tempModelCost + tempDataCost)  {  // if we have improved...
				NumSpec++;
				Spec[D[j].Idx] = 1;
				NumAccountedFor = tempNumAccountedFor;
				NumSpecSum = tempNumSpecSum;
				PosSpecSum = tempPosSpecSum;
				SpecCodeLength = tempSpecCodeLength;
				ModelCost = tempModelCost;
				DataCost = tempDataCost;
			}
			
			
			
		}  // closes loop over the canonial seqs in [a,b]
		
		TotalCostOverAllRows += ModelCost + DataCost + 1 + /* the "+1" is for the bit saying whether NumSpec==0 or is >0 */
					(log2Nab * (double)(NumSpec > 0))  ;  /* the last term here is for specifying
																		how many 1's are in each of the rows */
		
		/* increment the output parameters for the values in each row */
		*DataCostOutput += DataCost;
		*DeltaCost += NumSpecSum;
		*PCost += PosSpecSum;
		*NumDelts += NumSpec;
		*log2NabSum += log2Nab * (double)(NumSpec > 0);  /* add cost of the number of delta=1's for the case when we specify
														more than zero of them  */
		*FixedBitCost += 1.0;  /* this will just end up being the number of rows, but I'm feeling lazy and so will calculate by
									adding up 1.0's  */
									
									
		/***********  A SECTION FOR COMPUTING THE KULLBACK-LEIBLER DIVERGENCE BETWEEN P* AND q ************/
		/* first comute the P* for this row, do it as done above */
		prop = 0.0;
		for(i=0;i<ta->N;i++)  {
			if(Spec[i]==0)
				prop += Freqs[i];
		}
		prop /= ((double)B->NumSeq - NumAccountedFor)/(double)(B->NumSeq);
		prop = 1.0/prop;
		
		// then actually fill the pstar array:
		for(i=0;i<ta->N;i++)  {
			if(Spec[i]==0)
				pstar[i] = prop * Freqs[i];
			else 
				pstar[i] = (double)NumInTA[i]/(double)tz->Nd[r];
		}
		/* and compute the weight for this row:  */
		KLweight = (double)tz->Nd[r] / (double)tz->N;
		
		/* then compute the divergence (note that this is the symmetrical divergence) */
		KLD_r = 0.0;
		for(i=0;i<ta->N;i++)  {
			KLD_r += (pstar[i] - Freqs[i]) * log(pstar[i]/Freqs[i]);
		}
		*KLD += KLweight * KLD_r;
			
	}  // closes the loop over rows
	
	 
	free(Freqs);
	free(MLogFreqs);
	free(Zseq);
	free(Aseq);
	free(D);
	free(Spec);
	free(NumInTA);
	free(pstar);
	
	return(TotalCostOverAllRows);
	
}



void StoreTransInfo(findcan_struct **F, int N, trans_info_struct *T, int NumSeq)
{
	int i,j, x;
	
	T->N = N;
	
	// allocate space to the Nd and the first layer of Idx
	T->Nd = (int *)ECA_CALLOC( (size_t)T->N, sizeof(int));
	T->Idx = (int **)ECA_CALLOC( (size_t)T->N, sizeof(int *));
	
	// cycle over the canonicals
	for(i=0;i<T->N;i++)  {
		// count up the Nd
		T->Nd[i] = F[i]->RowSum;
		
		// then allocate to Idx's second layer:
		T->Idx[i] = (int *)ECA_CALLOC( (size_t)T->Nd[i], sizeof(int));
		
		// then fill Idx[i] in:
		x = 0;
		for(j=0;j<NumSeq;j++)  {
			if(F[i]->Row[j])
				T->Idx[i][x++] = j;
		}
		
	}
}





// does the calculations to compute the entropy, or at least to collate different types of 
// sequences for a block starting at a and ending at btop-1.  So, for filling in matrices,
// this should be called with a=a and btop = B->NumMarkers, and the flag FillinMats == 1  
// Sometimes, however, we just will want to collate the types...for a block [a,b].  In such
// cases, call this with a=a, btop = b+1 and FillinMats == 0.  
// note that memory gets allocated and things typically get initialized as if a==0 and btop = NumMarkers
void MoreEfficientEntropy(int a, MDL_struct *B, int btop, int FillinMats)
{
	int i,j;  // index for sequences and for markers within sequences, respectively.
	int NumNodes, Depth;
	int NumInNodes = B->NumSeq * B->NumMarkers * 2;
	bin_node *Root, *Current;

	// If it is null allocate memory to B->Nodes.  And set SameSeqs to NULL within each
	if(B->Nodes==NULL)  {
		B->Nodes = (bin_node *)ECA_CALLOC((size_t)(NumInNodes),sizeof(bin_node));
	}
	
	// allocate to the Sets
	if(B->S == NULL) {
		B->S = (set_struct *)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(set_struct ));
		for(i=0;i<B->NumMarkers;i++)  {
			B->S[i].NumInSet = (int *)ECA_CALLOC((size_t)B->NumSeq, sizeof(int));
		}
	}
	// and also allocate to the other sets needed there
	if(B->Sets == NULL)  {
		B->Sets = (int **)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int *));
		for(i=0;i<B->NumSeq;i++)  {
			B->Sets[i] = (int *)ECA_CALLOC((size_t)B->NumMarkers, sizeof(int));
		}
	}
	// and always initialize NumSets and NumInSet
	for(i=0;i<B->NumMarkers;i++)  {
		B->S[i].NumSets = 0;
		for(j=0;j<B->NumSeq;j++)  {
			B->S[i].NumInSet[j] = 0;
		}
	}
	
	// allocate to root
	Root = (bin_node *)ECA_MALLOC(sizeof(bin_node));
	
	
			
	// initialize all the nodes
	for(i=0;i<NumInNodes;i++)  {
		InitNode(&(B->Nodes[i]));
	}
	// then initialize root
	InitNode(Root);
	
	// then initialize our counters
	NumNodes = 0;
	
	
	// FIRST PASS THROUGH TO COUNT UP THE NUMBER OF TIPS AND THE VISITS
	// then cycle over the sequences
	for(i=0;i<B->NumSeq;i++)  {

		// at the beginning of each sequence, set Current to Root
		// and Depth to 0, and also prepare to start counting zeroes and ones
		Current = Root;
		Depth = 0;	
		// now, cycle over the markers
		for(j=a;j<btop;j++)  {
			// depending on if it is a zero or a one, follow that branch out of the current
			// node, assigning memory as necessary.
			if(Current->Br[B->Y[i][j]] == NULL)  {  // if we need to assign some memory here
				Current->Br[B->Y[i][j]] = &(B->Nodes[NumNodes++]);
				// then move our current pointer along to the node at the end of the appropriate branch
				Current = Current->Br[B->Y[i][j]];
				// record this as a visit
				Current->Visits++;
				// and set the Idx in it to the appropriate NumSets (which will get incremented in a few lines)
				Current->Idx = B->S[a+Depth].NumSets;
				// and increment the number of entries in that set
				B->S[a+Depth].NumInSet[Current->Idx] += 1;
				// Increment NumSets
				Current->Idx = B->S[a+Depth].NumSets++;
			}
			else {  // otherwise, the node has been traversed through before and it is simple
				Current = Current->Br[B->Y[i][j]];
				// don't forget to record this as a visit
				Current->Visits++;
				// and increment the appropriate number of entries in the set
				B->S[a+Depth].NumInSet[Current->Idx] += 1;
			}
			// always increment depth
			Depth++;
		}  // closes j-loop over markers
	} // closes i-loop over sequences
	// END OF FIRST PASS
	
	
	// now, allocate to the Sets elements
	for(i=a;i<btop;i++)  {
		B->S[i].Set = (int **)ECA_CALLOC((size_t)B->S[i].NumSets,sizeof(int **));
		for(j=0;j<B->S[i].NumSets;j++)  {
			B->S[i].Set[j] = (int *)ECA_CALLOC( (size_t)B->S[i].NumInSet[j], sizeof(int) );
		}
	}


	// SECOND PASS THROUGH TO FILL IN THE SETS
	// first initialize NumInSet another time
	for(i=0;i<B->NumMarkers;i++)  {
		for(j=0;j<B->NumSeq;j++)  {
			B->S[i].NumInSet[j] = 0;
		}
	}
	//  cycle over the sequences
	for(i=0;i<B->NumSeq;i++)  {
		// at the beginning of each sequence, set Current to Root
		// and Depth to 0
		Current = Root;
		Depth = 0;	
		// now, cycle over the markers
		for(j=a;j<btop;j++)  {
			// depending on if it is a zero or a one, follow that branch out of the current node to the next
			Current = Current->Br[B->Y[i][j]];	
			// record the sequence number in Sets (using the previously stored Idx in Current)
			B->S[a+Depth].Set[Current->Idx][  B->S[a+Depth].NumInSet[Current->Idx] ] = i;
			// then increment NumInSet
			B->S[a+Depth].NumInSet[Current->Idx]++;
			// and record the index of the Set for this sequence in Sets:
			B->Sets[i][a+Depth] = Current->Idx;
			// increment the Depth
			Depth++;
		}  // closes j-loop over markers
	} // closes i-loop over sequences
	

/*
out = fopen("Canon.txt","w");
fprintf(out,"a = %d\n\n",a);
for(i=a;i<B->NumMarkers;i++)  {
	fprintf(out,"\n");
	for(j=0;j<B->S[i].NumSets;j++)  {
		fprintf(out,"%d\t",B->S[i].Set[j][0]);
	}
}
fclose(out);
*/	
	
	
	// use the stuff in the array of sets to compute the E and SumOfCodeLengths arrays
//	StandardScoreFromSetsArray(a,B);
	
	if(FillinMats) {
		ComputeEntropyByCompats(a,B);
	}
	else {
		// here we have a function that finds  and prints canonicals
		;
		ReportCanonicalSeqs(a,B,btop-1, gCanonSeqFileName);
	}
	
	
	
	

	// free the memory that was allocated to Sets
	for(i=a;i<btop;i++)  {
		for(j=0;j<B->S[i].NumSets;j++)  {
			free(B->S[i].Set[j]);
		}
		free(B->S[i].Set);
		
	}
	// and free root
	free(Root);
	
}

// this fills in all the matrices that need to be filled
// in to implement the DP algorithm.  
void MoreEfficientFillInMatrices(MDL_struct *B)
{
	int a,i,j;
	
	// check to see if space has been allocated to the matrices.  If not, do so.
	if(B->E == NULL)
		B->E = AllocEntMat(B->NumMarkers);
	if(B->CanonSeqMultParCost == NULL)
		B->CanonSeqMultParCost = AllocEntMat(B->NumMarkers);
	if(B->ChosenQPrior==NULL)  {
		B->ChosenQPrior = (enum qtype_enum **)ECA_CALLOC((size_t)B->NumMarkers,sizeof(enum qtype_enum *));
		for(j=0;j<B->NumMarkers;j++)  {
			B->ChosenQPrior[j] = (enum qtype_enum *)ECA_CALLOC( (size_t)(j+1), sizeof(enum qtype_enum));
		}
	}
	if(B->SumOfCodeLengths == NULL)
		B->SumOfCodeLengths = AllocEntMat(B->NumMarkers);
	if(B->NT == NULL)
		B->NT = AllocNTMat(B->NumMarkers);
	if(B->Tr==NULL)  {
		B->Tr = (trans_info_struct **)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(trans_info_struct *));
		for(i=0;i<B->NumMarkers;i++)  {
			B->Tr[i] = (trans_info_struct *)ECA_CALLOC( (size_t)(i+1), sizeof(trans_info_struct));
		}
	}
	else {
		printf("\nBad news here.  Trans Info Struct already allocated to.  Quitting!!\n\n");
		exit(1);
	}
	
		
	// initialize both E and NT to -1's
	InitEntMat(B->E, B->NumMarkers);
	InitEntMat(B->CanonSeqMultParCost, B->NumMarkers);
	InitEntMat(B->SumOfCodeLengths, B->NumMarkers);
	InitNTMat(B->NT, B->NumMarkers);
	
	// and compute the code lengths from the filled data:
	//ComputeCodeLengths(B, B->Y);
	
	// then cycle over the starting points of blocks
	for(a=0;a<B->NumMarkers;a++)  {
		
		if(gVerboseFlag)  {
			printf("\nComputing Matrices, a = %d",a);
			fflush(stdout);
		}
		// for each starting point, compute the entropy and such internally
		MoreEfficientEntropy(a,B,B->NumMarkers,1); 
		
	}

}





// fill the E matrix as it should be filled given an array of sets.
// also fills up the SumOfCodeLengths array
void StandardScoreFromSetsArray(int a, MDL_struct *B)
{	
	int b;
	int i,k;  // index for sequences and for markers within sequences, respectively.
	double Entropy, Totals, temp_p;
	double SumOfCodeLengths, TipCodeLength;
	double dtemp;

	// cycle over all the blocks of the form [a,b]
	for(b=a;b<B->NumMarkers;b++)  {
	
	// cycle over all the types and their sequences to compute the sum of their
	// Code Length:
	SumOfCodeLengths = 0.0;
	for(i=0;i<B->S[b].NumSets;i++)  { 
		TipCodeLength = 0.0;
		for(k=0;k<b-a+1;k++)  { 
				// this is sort of tricky.  We sum up the lengths corresponding to the 0's and 1's of 
				// any of the sequences in the SameSeq set for this particular tip
				TipCodeLength += B->CodeLengths[k+a][SAME_SEQ_VAL(B,a,b,i,k) ];
		}
//printf("\n%.5f", TipCodeLength);
		SumOfCodeLengths += TipCodeLength;
	}
	
	
	// and assign that to the proper place:
	B->SumOfCodeLengths[b][a] = SumOfCodeLengths;
	
	// then compute the entropy, using log base 2.
	// first compute the denominator
	Totals = 0.0;
	Entropy = 0.0;
	for(i=0;i<B->S[b].NumSets;i++)  {  
		Totals += (double)B->S[b].NumInSet[i];
	}
	for(i=0;i<B->S[b].NumSets;i++)  {  
		temp_p = (double)B->S[b].NumInSet[i]/Totals;
		dtemp = -temp_p * log(temp_p) / log(2.0);
		Entropy -= temp_p * log(temp_p) / log(2.0);
	}
	// and record it in the E array within B
	B->E[b][a] = Entropy;
	// and record the number of types too
	B->NT[b][a] = B->TipsNums[b];
	
	}
}



// this takes all the canonical sequences and checks to see which of the 
// data sequences are compatible with them when you ignore the missing data
void CompileCompatibleSeqs(int a, MDL_struct *B)
{
	int i,j,b;
	
	// for temporary usage
	int *CurrSwitch = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));  // to hold the list of the
																			// canonical seqs compatible
																			// at the current b
	int *PrevSwitch = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int)); // to hold the list of compatibles at the previous b

	int **tempCoSeqs = (int **)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int *));
	
	for(i=0;i<B->NumSeq;i++)  {
		tempCoSeqs[i] = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
	}
	
	// first allocate memory if necessary:
	if(B->Canon==NULL)  {
		B->Canon = (canon_container *)ECA_CALLOC( (size_t)B->NumMarkers, sizeof(canon_container));
		for(i=0;i<B->NumMarkers;i++)  {
			// allocate memory (quite a bit more than needed, actually) to the CS field of each
			B->Canon[i].CS = (compat_struct *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(compat_struct));
		}
	}
	// then fill in the elements of the canon container. 
	// first part is easy -- just fill in the number of sets and copy over the canonical seq indexes.
	// while we are at it, let's initialize some values
	for(i=a;i<B->NumMarkers;i++)  {
		B->Canon[i].N = B->S[i].NumSets;
		for(j=0;j<B->Canon[i].N;j++)  {
			B->Canon[i].CS[j].Idx = B->S[i].Set[j][0];
			B->Canon[i].CS[j].nC = 0;
		}
	}
	
	
	// then cycle over the different seqs and for each, cycle over the b's and determine which 
	// canonical seqs they are compatible with
	for(i=0;i<B->NumSeq;i++)  {
		
		for(b=a;b<a+1/*B->NumMarkers*/;b++)  {  // cycle over the markers (block endpoints)
		// step one, initialize things using the b=a case
		if(b==a)  {
			for(j=0;j<B->Canon[b].N;j++) {
				// then determine whether the current sequence is compatible with that or not
				if( B->Y[B->Canon[b].CS[j].Idx][b] == B->Yobs[i][b]  ||  B->Yobs[i][b] < 0 )
					CurrSwitch[j] = 1;
				else 
					CurrSwitch[j] = 0;
			}
		}
		else {  // otherwise, you have to use PrevSwitch and the old seqs to determine CurrSwitch
		
		}
		// and use the 0's and 1's in CurrSwitch to increment nC and the list of CoSeqs in each CS
		for(j=0;j<B->Canon[b].N;j++) {
			if(CurrSwitch[j]) {
				tempCoSeqs[j][B->Canon[b].CS[j].nC++] = i;
			}
		}
		
		
		// allocate memory to and transfer across the CoSeqs
		 
/*		
		PrevN = 0;
				
		for(b=a;b<B->NumMarkers;b++)  {  // cycle over the markers (block endpoints)
			pidx = 0;
			CurrN = 0;
			for(j=0;j<B->Canon[b].N;j++)  {  // cycle over the canonical sequences at b
				if(pidx < PrevN) {  // only be doing this if pidx hasn't run off the end
					if(B->Canon[b].CS[j].Idx == B->Canon[b-1].CS[pidx].Idx )  {
						CurrSwitch[j] = PrevSwitch[pidx];
						pidx++;
					}
					else 
						CurrSwitch[j] = 2;
						pidx++;
				}
				else {
					CurrSwitch[j] = 2;
				}
				CurrN++;
			}
			// down here do whatever one does with the CurrSwitch array.  Basically you have the chance to
			// turn 2's into 1's or 0's and the chance to turn 1's into 0's, as appropriate.
			
			// then make the new assignments to counters and things
			PrevSwitch = CurrSwitch;
			PrevN = CurrN;
			
			
			
		}
*/	
	}   // closes the loop over b
	}   // closes the loop over i (sequences)

	free(CurrSwitch);																	
	free(PrevSwitch);
	for(i=0;i<B->NumSeq;i++)  {
		free(tempCoSeqs[i]);
	}
	free(tempCoSeqs);

}


// initialize the values in the findcan struct to accomodate dealing 
// with a block in which there are N canonical sequences
void InitFindCan(int N, int b, findcan_struct **FC, MDL_struct *B)
{
	int i,j;	
	
	for(i=0;i<N;i++)  {
		FC[i]->RowSum = 0;
		FC[i]->Idx = B->S[b].Set[i][0];
		for(j=0;j<B->NumSeq;j++)  {
			FC[i]->Row[j] = 0;
		}
	}
}



int **AllocAndInitCompatSwitches(int NumSeq)
{
	int i,j;
	int **temp;
	
	temp = (int **)ECA_CALLOC( (size_t)NumSeq, sizeof(int *));
	for(i=0;i<NumSeq;i++)  {
		temp[i] = (int *)ECA_CALLOC( (size_t)NumSeq, sizeof(int));
		for(j=0;j<NumSeq;j++)  {
			temp[i][j] = 2;   // 2 means that the sequence has not been yet checked for compatibility to sequence j
		}
	}
	return(temp);
}

void FreeCompatSwitches(int **temp, int NumSeq)
{
	int i;
	
	for(i=0;i<NumSeq;i++)  {
		free(temp[i]);
	}
	free(temp);

}


// returns a one if seq i is compatible with seq Idx for the block from
// a to b inclusive.  Uses CS as compatibility switches to know if it 
// needs to check all sites from a to b (if CS[Idx] == -1) or if it can
// just check the b-th site (if CS[Idx] == 1), or if it can proclaim
// it incompatible immediately (if CS[Idx] == 0).  This updates 
// CS[Idx] while it is at it.  
int IsCompatible(int i, int a, int b, int *CS, int Idx, MDL_struct *B)
{
	int j;
	
	switch(CS[Idx]) {
		case(0):
			return(0);
			break;
		
		case(1):  // if it has been compatible up till this point, check compatibility at b
			if(B->Yobs[i][b] == B->Y[Idx][b] || B->Yobs[i][b] < 0)  {  // if it is equal or the data in i is missing
				CS[Idx] = 1;
				return(1);
			}
			else {
				CS[Idx] = 0;
				return(0);
			}
			printf("\n\nWarning!  Unexpected result in IsCompatible() with CS[Idx] == 1\n\n");
			break;
		
		case(2):  // in this case we have to check for compatibility all the way from a to b
			for(j=a;j<=b;j++)  {
				if(B->Yobs[i][j] >= 0 && B->Yobs[i][j] != B->Y[Idx][j])  {  // if there is a mismatch
					CS[Idx] = 0;
					return(0);
				}
			}
			// if we get to here, it was compatible:
			CS[Idx] = 1;
			return(1);
			break;
		default:
			printf("\n\nBummer Dude!  You reached the default in the switch in IsCompatible()\n\n");
			exit(1);
	}
	return(-55);
}


int CompareFindCanRows(const void *vT1, const void *vT2)
{
	int value = 0;
	
	const findcan_struct **T1p = (const findcan_struct **)vT1;
	const findcan_struct **T2p = (const findcan_struct **)vT2;
	
	const findcan_struct *T1 = *(T1p);
	const findcan_struct *T2 = *(T2p);
	
	if(T1->RowSum < T2->RowSum) 
		value = 1;
	if(T1->RowSum > T2->RowSum)
		value = -1;
		
	return(value);
}


// given a filled out findcan_struct, for N canonical sequences, 
// this function greedily chooses which canonicals
// should be chosen.
// Currently the compare function does not take account of ties very appropriately.
int SortAndChooseCanonicals(findcan_struct **FC, MDL_struct *B, int N)
{
	int i,j,k=0, SumOfRowSums=1;
	findcan_struct **FCP;
	
	
	
	while(SumOfRowSums>0 && k<N) {
		
		// sort the array from k to N in order of decreasing RowSums
		FCP = &(B->FC[k]);  // make a pointer to the correct starting place in the array of pointers to findcans
		qsort((void *)FCP,(size_t)(N-k), sizeof(findcan_struct *),CompareFindCanRows);
		
		// now, the k-th element is the current "new member" to the chosen set
		// so anywhere it has a 1 in its Row array, we have to put a zero in all other
		// Rows that are not yet in the chosen set
		for(i=0;i<B->NumSeq;i++)  {  // cycle over the elements of Row
			if(B->FC[k]->Row[i]==1)  {
				for(j=k+1;j<N;j++) {
					if(B->FC[j]->Row[i]==1) {
						B->FC[j]->Row[i] = 0;
						B->FC[j]->RowSum--;
					}
				}
			}
		}
		
		// increment k
		k++;
		// compute SumOfRowSums
		SumOfRowSums = 0;
		for(j=k;j<N;j++)  {
			SumOfRowSums += FC[j]->RowSum;
		}
	}
	
	// now that we are done, FC has been sorted so that the first k elements represent
	// the chosen canonicals
	return(k);
}


// returns the entropy, given a sorted set of N chosen canonical seqs in FC
double EntropyOfSortedCanonicals(findcan_struct **FC, int N)
{
	int i;
	double p,Total=0.0,Entropy=0.0;
	
	// cycle once to get the total
	for(i=0;i<N;i++)  {
		Total += (double)FC[i]->RowSum;
	}
	
	// cycle again to get the entropy
	for(i=0;i<N;i++)  {
		p = (double)FC[i]->RowSum/Total;
		Entropy -= p * log(p)/log(2.0);
	}
	
	return(Entropy);
}

// returns the cost of encoding the multinomial model parameters for the 
// sorted canonicals.  Does not include the "number of categories" cost
double ComputeCanonSeqMultParCost(findcan_struct **FC, int N, int S, enum qtype_enum *Q) 
{
	int j,s;
	double cCost=0.0,eCost=0.0,oCost=0.0,Cost=0.0;
	int cTotal=0, NdivKCeil;
	/* note eCost is for Ewens, o is for occupancy, c is for standard */
	int *C = (int *)ECA_CALLOC((size_t)N, sizeof(int));
	
	/*  fill the C array.  Note that it is initialized to 0's by calloc.
		This array will get used in both OCCUPANCY and EWENS */
	if(gQPrior==OCCUPANCY || gQPrior==EWENS || gQPrior==MIXTURE) {
		for(s=0;s<S;s++)  {
			C[FC[s]->RowSum]++;
		}
	}
	
	
	if(gQPrior==STANDARD || gQPrior==MIXTURE)  {
		/* default way of doing it that favors skewed frequency distribution of types */
		// cycle once to get the total
		
		for(s=0;s<S;s++)  {
			cTotal += FC[s]->RowSum;
		}
		
		// cycle again to compute the code lengths
		for(s=0;s<S-1;s++)  {  /* only cycle to < S-1 because the last frequency is determined
								 'cuz they all have to sum to one */  
		
			/* compute \ceil{N/S^{(k)}} */
			NdivKCeil = (cTotal / (S-s)) + ( (cTotal % (S-s)) != 0);
			
			cCost  += log(cTotal - (S-s) - NdivKCeil + 2)/log(2.0);
			cTotal -= FC[s]->RowSum;
		}
	}
	
	if(gQPrior==OCCUPANCY || gQPrior==MIXTURE) {
		/** new way of doing it, as per the manuscript:
			Variables here correspond to notation in the manuscript as follows:
				N 					-->		N    (number of sequences )
				S					-->		S_k  (number of types)
				FC[s]->RowSum		-->		X_s^{(k)}   (number of seqs of type s)
				C[j]				-->   	C_j			(number of types represented by j sequences
														 in the sample)
		*/
		
		
		
		/* now start computing cost.  Compute all the factorials in terms of natural log of the
		factorials, then divide by log(2.0) at the end, and that will do it all */
		
		/* First term */
		oCost += LogFactorial(N);
		for(s=0;s<S;s++)  {
			oCost -= LogFactorial(FC[s]->RowSum);
		}
		
		/* Second term */
		oCost += LogFactorial(S);
		for(j=0;j<N;j++)  {
			if(C[j]) {  /* if C[j] is not zero.  (do this just to be a little
							 faster than taking the LogFactorial of 0 all the time.  This
							 could probably be done much more efficiently---especially for large N */
				oCost -= LogFactorial(C[j]);
			}
		}
		
		/* Third term */
		oCost += LogFactorial(N-S) + LogFactorial(S-1) - LogFactorial(N-1);
		
		
		/* make it log base 2 */
		oCost /= log(2.0);
	}
	
	if(gQPrior==EWENS || gQPrior==MIXTURE) {
		/** Using the multivariate Ewens distribution conditional on S_k:
			Variables here correspond to notation in the manuscript as follows:
				N 					-->		N    (number of sequences )
				S					-->		S_k  (number of types)
				FC[s]->RowSum		-->		X_s^{(k)}   (number of seqs of type s)
		*/
		
		/* Computing cost following equation 41.8 in Johnson, Kotz, and Balakrishnan's
			"Multivariate Discrete Distributions" */
		
		eCost = LogFactorial(N);
		eCost -= LogFactorial(S);
		eCost -= RecursiveLogStirling3(&gStirlingArray,N,S,N);
		for(s=0;s<S;s++)  {
			eCost -= log(FC[s]->RowSum);
		}
		
		/* now we increase this probability a little bit to account for the fact that we
		   have ordered these types.  This is the S!/ \prod C[j]! term.  I could have divided
		   out the S! terms from the get go, but it's a little clearer to see what is going on
		   here since I have left it in.  */
		eCost += LogFactorial(S);
		
		for(j=0;j<N;j++)  {
			if(C[j]) {  /* if C[j] is not zero.  (do this just to be a little
							 faster than taking the LogFactorial of 0 all the time.  This
							 could probably be done much more efficiently---especially for large N */
				eCost -= LogFactorial(C[j]);
			}
		}
		
		
		/* make it log base 2, and multiply it by -1 since we have the log of the
		 probability  */
		eCost /= -log(2.0);
	
	}
	
	/*  now choose the minimum of the ones that have been computed */
	*Q = gQPrior;
	switch(gQPrior) {
		case(STANDARD):
			Cost = cCost;
			break;
		case(EWENS):
			Cost = eCost;
			break;
		case(OCCUPANCY):
			Cost = oCost;
			break;
		case(MIXTURE):
			Cost = cCost;
			*Q = STANDARD;
			if(eCost<Cost)  {
				Cost = eCost;
				*Q = EWENS;
			}
			if(oCost < Cost)  {
				Cost = oCost;
				*Q = OCCUPANCY;
			}
			break;
		default:
			fprintf(stderr,"\n\nWarning! Reached default in switch() in ComputeCanonSeqMultParCost()\n\n");
	}
	
	
	free(C);
	return(Cost);

}

// returns the sum of the code lengths for sending the chosen canonicals
double SumCodeLengthOfSortedCanonicals(findcan_struct **FC, int N, int a, int b, MDL_struct *B)
{
	int i,j,NumOnes;
	double p,q;
	double CodeLength, SumOfCodeLengths=0.0;
	
	for(j=a;j<=b;j++)  {  // cycle over the sites
		NumOnes = 0;
		// cycle once over the chosen canonical seqs to count ones amongst the chosen canonicals
		for(i=0;i<N;i++)  {
			NumOnes += B->Y[ FC[i]->Idx ][j];
		}
		p = (double)NumOnes / (double)N;
		q = (double)(N - NumOnes)/(double)N;
		
		
		// MAJOR ECA OVERHAUL:  3/27/03.  I think the program crashes on monomorphic sites because the
		// code length has a log(0.0) in it.  Of course, if either p or q is equal to zero then the code
		// length will be zero for that column, cuz it carries no information really.
		if(NumOnes != N && NumOnes != 0)
			CodeLength = -(double)NumOnes * log(p)/log(2.0)  - (double)(N - NumOnes) * log(q)/log(2.0);
		else
			CodeLength = 0.0;
			
		SumOfCodeLengths += CodeLength;
		
		// for each column, we also include the cost of sending the estimated value of p
		// (which we do to specify the binomial model)
		if(N>1)  // this should always be the case
			SumOfCodeLengths += log(N-1)/log(2.0);
		else
			SumOfCodeLengths += log(N)/log(2.0);
		
	}
	
	return(SumOfCodeLengths);
}




/* 
	given N canonical sequences, and ordered in the FC struct, this
	function puts the number of mismatches between those sequences (pairwise
	comparisons) into the matrix Mis for the block from a to b (where a to a
	was when Mis was initialized, and we build on that for increasing b).
	
	This operates on the N canonical sequences rather than the NumChosen because
	we can't afford to occasionally miss one that didn't get in the Chosen set but
	which might reappear later in the chosen set as we cycle over b.
	
	Note also that when we change anything in Mis, we want to make sure to change
	both [i][j] and [j][i] to keep it symmetric---that will make all the book-keeping
	easier, I think.
*/
void IncrementMismatchMatrix(int N, findcan_struct **FC, int **Mis, int a, int b, MDL_struct *B)
{
	int i,j,k;
	
	for(i=0;i<N;i++)  {
		for(j=i+1;j<N;j++)  {
			if(Mis[FC[i]->Idx][FC[j]->Idx] > -1)  {  // if this is the case, then the pair has been compared already
								// and we just need to add 0 or 1 to the value, depending on 
								// how the b-th element compares
				if(B->Y[FC[i]->Idx][b] != B->Y[FC[j]->Idx][b])  {  // if they mismatch add a one
					Mis[FC[i]->Idx][FC[j]->Idx]++;
					Mis[FC[j]->Idx][FC[i]->Idx]++;
				}
				// otherwise do nothing
			}
			else  {  // if there still is a -1 in this entry, then we have to add up the number
					 // missing from a to b, inclusive and store that in there
				Mis[FC[i]->Idx][FC[j]->Idx] = 0;
				Mis[FC[j]->Idx][FC[i]->Idx] = 0;  // initialize to accumulate a sum
				// then cycle over k from a to b and count up the mismatches
				for(k=a;k<=b;k++)  {
					if(B->Y[FC[i]->Idx][k] != B->Y[FC[j]->Idx][k]) {
						Mis[FC[i]->Idx][FC[j]->Idx]++;
						Mis[FC[j]->Idx][FC[i]->Idx]++;
					}
				}
			}
		}
	}
	
}


/*
	given NC chosen canonicals, ordered in FC, and having numbers of mismatching sites
	between them as in Mis, then compute the code length required to encode them by
	a greedy diff-matching method.  This method is not currently used.

*/
void ComputeCodeLengthsUsingDiffs(int NC, findcan_struct **FC, int **Mis, int a, int b, MDL_struct *B)
{
	double tempCL;  // temporary code length of a sequence
	int i,j,k;
	int delta;
	double TotalCL = 0.0;
	double Q;  // the sum of the descending logs part
	int l = (b-a+1);
	
	for(j=0;j<NC;j++)  {  // cycle over the chosen canonical seqs
		// compute the codelength
		tempCL = 0.0;
		for(k=a;k<=b;k++)  {
			tempCL += B->CodeLengths[k][B->Y[FC[j]->Idx][k]];
		}
		// if this was the first canonical sequence, then that is the
		// codelength. otherwise we will allow the possibility that we can
		// encode it more cheaply by encoding diffs.
		if(j>0) {
			// now find the minimum number of mismatches between it and any previously occuring
			// sequence.  Store this in the variable called delta.
			delta = b-a+1;
			for(i=0;i<j;i++)  {
				if(Mis[FC[i]->Idx][FC[j]->Idx] < delta) {
					delta = Mis[FC[i]->Idx][FC[j]->Idx];
				}
			}
			// now, with delta in hand, we can compute the sum of the logs part
			Q=0.0;
			for(k=0;k<delta;k++)  {
				Q += log(l-k)/log(2.0);
			}
			// and now check to see if that is less than tempCL.  if it is, then
			// assign it to tempCL
			if(Q<tempCL)  {
				tempCL = Q;
			}
			
		}
		// now, add log2(j+1) to tempCL and add that to the TotalCL
		tempCL += log(j+1)/log(2.0);
		TotalCL += tempCL;
	}
	// and assign to the SumOfCodeLengths
	B->SumOfCodeLengths[b][a] = TotalCL;
}


void ComputeEntropyByCompats(int a, MDL_struct *B)
{
	int i,b,j,N,ii,jj;
	int **CompatSwitches;
	FILE *out;
	int Sum;
	int verb = 0;
	int NumChosen;
	int **MismatchNums;  
	
	if(B->FC==NULL)  {  // allocate if necessary
		B->FC = (findcan_struct **)ECA_CALLOC( (size_t)B->NumSeq, sizeof(findcan_struct *));
		for(i=0;i<B->NumSeq;i++)  {
			B->FC[i] = (findcan_struct *)ECA_MALLOC( sizeof(findcan_struct));
			B->FC[i]->Row = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
		}
		if(verb)  {
			out = fopen("EntByCompats.test","w");
			fclose(out);
		}
	}
	
	
	
	// allocate to the compat switches and initialize them:
	CompatSwitches = AllocAndInitCompatSwitches(B->NumSeq);
	
	// then allocate to MismatchNums, which is of the same dimension,
	// however, we must initialize it to -1's
	MismatchNums = AllocAndInitCompatSwitches(B->NumSeq);
	for(i=0;i<B->NumSeq;i++)  {
		for(j=0;j<B->NumSeq;j++)  {
			MismatchNums[i][j] = -1;
		}
	}
	if(verb) {	
		out = fopen("EntByCompats.test","a");	
	}
	
	for(b=a;b<B->NumMarkers;b++)  {  // cycle over the block boundaries to fill the FC struct
	
		// set the value of N, (easier to use than the full expression each time)
		N = B->S[b].NumSets;
		// initialize the FC
		InitFindCan(N,b,B->FC,B);
		
		for(i=0;i<B->NumSeq;i++)  {  // cycle over the sequences
			for(j=0;j<N;j++)  {  // cycle over the sets of canonical sequences
				B->FC[j]->Row[i] = IsCompatible(i, a, b, CompatSwitches[i], B->FC[j]->Idx,B);
				B->FC[j]->RowSum += B->FC[j]->Row[i];
			}
		}
		
		
		NumChosen = SortAndChooseCanonicals(B->FC,B,N);
		
		
		
		// now that those things are sorted we can compute the number of mismatches
		// between the different canonical sequences
		IncrementMismatchMatrix(N, B->FC,MismatchNums,a,b,B);
		
		// here we assign the entropy:
		B->E[b][a] = EntropyOfSortedCanonicals(B->FC, NumChosen);
		
		// here we store the trans_info information
		StoreTransInfo(B->FC, NumChosen, &(B->Tr[b][a]), B->NumSeq);
		
		// here we assign the cost of encoding the multinomial model---for this we send the
		// number of categories (NumChosen) as if it were uniform on the integers between 0 and
		// log_2(NumSeq), then we send the cost of encoding the estimated frequencies of those 
		// canonicals
		B->CanonSeqMultParCost[b][a] = ComputeCanonSeqMultParCost(B->FC, B->NumSeq, NumChosen, &B->ChosenQPrior[b][a]);
		B->CanonSeqMultParCost[b][a] += log(B->NumSeq) / log(2.0);  // for the "number of categories"
		
		// here we assign the sum of codelengths
		B->SumOfCodeLengths[b][a] = SumCodeLengthOfSortedCanonicals(B->FC,NumChosen,a,b,B);
		
		//ComputeCodeLengthsUsingDiffs(NumChosen,B->FC,MismatchNums,a,b,B);
		
		
		// and here we record the number of canonical sequences, too
		B->NT[b][a] = NumChosen;
		
		/**********************************************/
		if(verb)  {
			fprintf(out,"\n\na=%d  b=%d NumChosen = %d",a,b,NumChosen);
			for(jj=0;jj<N;jj++)  {  // cycling to print canonical sequences
				fprintf(out,"\n\t%d. %d:  ",jj,B->FC[jj]->Idx);
				for(ii=a;ii<=b;ii++)  {
					fprintf(out,"%d",B->Y[ B->FC[jj]->Idx ][ii]);
				}
			}	
			// then print the mismatch matrix:
			fprintf(out,"\nMismatchMatrix:");
			for(jj=0;jj<N;jj++)  { 
				fprintf(out,"\n\t");
				for(ii=0;ii<N;ii++)  {
					fprintf(out,"%d\t",MismatchNums[B->FC[jj]->Idx][B->FC[ii]->Idx]);
				}
			}	
			fprintf(out,"\nCompatMatrix:");
			Sum = 0;
			for(jj=0;jj<N;jj++)  {  // cycling to print compat matrix
				fprintf(out,"\n\t%d. %d (%d):  ",jj,B->FC[jj]->Idx,B->FC[jj]->RowSum);
				Sum += B->FC[jj]->RowSum;
				for(ii=0;ii<B->NumSeq;ii++)  {
					fprintf(out,"%d",B->FC[jj]->Row[ii]);
				}
			}	
		
			fprintf(out,"\n\t(%d)",Sum);
		}
		/************************************************/



		
	}  // closes the loop over b
	if(verb) {
		fclose(out);
	}

	FreeCompatSwitches(CompatSwitches,B->NumSeq);
	FreeCompatSwitches(MismatchNums,B->NumSeq);
	
}


/*  after finding the boundaries, you can call this, moving serially along
    the boundaries to print out the canonical sequences, etc.  Currently
    this also employs the variable FCP to be able to print out the  transition
    rates between the canonical sequences.  It's sort of silly the way I have set it
    up.  Notice that if you call this successively on blocks that are not in order, the transition
    rates are silly.  
    
 
*/
void ReportCanonicalSeqs(int a, MDL_struct *B, int b, char *FileName)
{
	int i,j,N,ii,jj,k;
	int **CompatSwitches;
	FILE *out;

	int NumChosen,maxi;
	char TempStr[100];
	double temp, tempNoMiss;
	int *NumOfTypesWNoMissingPre = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int)); /* the number of types having no missing data in the previous block */
	int *NumOfTypesWNoMissingCur = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int)); /* the number of types having no missing data in the current block */
	int *NumOfTypesWNoMissingBoth = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int)); /* the number of types having no missing data
								 in both the previous block and the current block */
	
	
	if(B->FC==NULL)  {  // allocate if necessary
		B->FC = (findcan_struct **)ECA_CALLOC( (size_t)B->NumSeq, sizeof(findcan_struct *));
		for(i=0;i<B->NumSeq;i++)  {
			B->FC[i] = (findcan_struct *)ECA_MALLOC( sizeof(findcan_struct));
			B->FC[i]->Row = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
		}
		
	}
	
	if(B->FCP==NULL)  {  // allocate if necessary
		B->FCP = (findcan_struct **)ECA_CALLOC( (size_t)B->NumSeq, sizeof(findcan_struct *));
		for(i=0;i<B->NumSeq;i++)  {
			B->FCP[i] = (findcan_struct *)ECA_MALLOC( sizeof(findcan_struct));
			B->FCP[i]->Row = (int *)ECA_CALLOC( (size_t)B->NumSeq, sizeof(int));
		}
	}
	
	// allocate to the compat switches and initialize them:
	CompatSwitches = AllocAndInitCompatSwitches(B->NumSeq);
	
	// open the file:
	sprintf(TempStr,"%s",FileName);
	out = fopen(TempStr,"a");
	

	// set the value of N, (easier to use than the full expression each time)
	N = B->S[b].NumSets;
	// initialize the FC
	InitFindCan(N,b,B->FC,B);
	
	for(i=0;i<B->NumSeq;i++)  {  // cycle over the sequences
		for(j=0;j<N;j++)  {  // cycle over the sets of canonical sequences
			B->FC[j]->Row[i] = IsCompatible(i, a, b, CompatSwitches[i], B->FC[j]->Idx,B);
			B->FC[j]->RowSum += B->FC[j]->Row[i];
		}
	}
		
		
	NumChosen = SortAndChooseCanonicals(B->FC,B,N); 
	
	/* here we compute the number of particular types in the current block that have no missing data */
/*	for(k=0;k<B->NumSeq;k++)  {
		for(i=0;i<NumChosen;i++) {
			if(B->FC[i]->Row[k]==1  && NoMissingData(B,k,a,b) ) 
				NumOfTypesWNoMissingCur[i]++;
		}
	}
*/
	
	/*  Then, if FCP has been set (i.e. we are up to our second interval) then
	 	compute and output the transition rates  */
	if(B->FCP_filled) {
		fprintf(out,"\n\nTransition rates from current (rows) to next (columns) group:\n\n");
		/* print out the column headings */
		fprintf(out,"        ");
		for(i=0;i<NumChosen; i++)  fprintf(out,"%10d_(%d) ",i+1,B->FC[i]->RowSum);
		
		/* then cycle over the rows  */
		fprintf(out,"\n");
		for(maxi=0;B->FCP[maxi]->RowSum>0;maxi++) ;
		for(i=0;B->FCP[i]->RowSum>0;i++)  {
		
			/* here we will compute the number of those types in the previous block that 
			   have no missing data in either the previous or both blocks */
	/*		for(k=0;k<B->NumSeq;k++)  {
				if(B->FCP[i]->Row[k]==1  && NoMissingData(B,k,gPrevReportedA,a-1) )  {
					NumOfTypesWNoMissingPre[i]++;
					if(NoMissingData(B,k,a,b) )
						NumOfTypesWNoMissingBoth[i]++;
				}
			}
	*/		
			
			
			
			fprintf(out, "\n%*d_(%d)",(int)log10(B->FCP[0]->RowSum) - (int)log10(B->FCP[i]->RowSum) +
										(int)log10(maxi) - (int)log10(i+1),
										i+1,B->FCP[i]->RowSum
										/*NumOfTypesWNoMissingBoth[i]*/);
			
			/* count the total number of 			*/		
										
			/*  and here cycle over the seqs that they're going to and compute the transition rates  */
			for(j=0;j<NumChosen;j++)  {
				temp = tempNoMiss = 0.0;
				for(k=0;k<B->NumSeq;k++)  {
					temp += (double)(B->FCP[i]->Row[k] * B->FC[j]->Row[k]);
				/*	tempNoMiss += (double)(B->FCP[i]->Row[k] * B->FC[j]->Row[k] *
										NoMissingData(B,k,gPrevReportedA,a-1) * NoMissingData(B,k,a,b) );*/
				}
				/*if(tempNoMiss>0.0)
					tempNoMiss /= (double)NumOfTypesWNoMissingBoth[i]; */
				fprintf(out,"%*.4f",15 + (int)log10(B->FC[j]->RowSum), temp/(double)B->FCP[i]->RowSum
								/*tempNoMiss*/);
			}
		}
	}
	
	/* print out the canonical seqs  */
	fprintf(out,"\n\na=%d  b=%d NumChosen = %d  Entropy = %.5f  SumOfCodes = %.5f",a,b,NumChosen,
					B->E[b][a], B->SumOfCodeLengths[b][a]);
	for(jj=0;jj<NumChosen;jj++)  {  // cycling to print canonical sequences
		fprintf(out,"\n\t%d_(%d):  ",jj+1,B->FC[jj]->RowSum);
		for(ii=a;ii<=b;ii++)  {
			if(gDalyBases==NULL)
				fprintf(out,"%d",B->Y[ B->FC[jj]->Idx ][ii]);
			else 
				fprintf(out,"%c",gDalyBases[ii][B->Y[ B->FC[jj]->Idx ][ii]]);
		}
	}	
	
	/* then transfer the important values from FC to FCP, and make the rest of the row sums
	 zeroes in it  */
	for(i=0;i<NumChosen;i++)  {
		B->FCP[i]->RowSum = B->FC[i]->RowSum;
		for(j=0;j<B->NumSeq;j++)  {
			B->FCP[i]->Row[j] = B->FC[i]->Row[j];
		}
	}
	for(i=NumChosen;i<B->NumSeq;i++)  B->FCP[i]->RowSum = 0;
	B->FCP_filled = 1;  /* let it be known this thing has been filled so it gets printed next time */

/*
	fprintf(out,"\nCompatMatrix:");
	Sum = 0;
	for(jj=0;jj<N;jj++)  {  // cycling to print compat matrix
		fprintf(out,"\n\t%d. %d (%d):  ",jj,B->FC[jj]->Idx,B->FC[jj]->RowSum);
		Sum += B->FC[jj]->RowSum;
		for(ii=0;ii<B->NumSeq;ii++)  {
			fprintf(out,"%d",B->FC[jj]->Row[ii]);
		}
	}	
	fprintf(out,"\n\t(%d)",Sum);
*/

	fclose(out);

	FreeCompatSwitches(CompatSwitches,B->NumSeq);
	free(NumOfTypesWNoMissingPre);
	free(NumOfTypesWNoMissingCur);
	free(NumOfTypesWNoMissingBoth);
		
	// set this for the next time:
	gPrevReportedA = a;
}


/* returns a 1 if there is no missing data (nothing smaller than 0's and 1's
 at the block extending from a to b inclusive in sequence "theSeq"  */
int NoMissingData(MDL_struct *B, int theSeq, int a, int b)
{
	int i,boing = 1;
	
	for(i=a;i<=b;i++)  {
		if(B->Yobs[theSeq][i] < 0 ) {
			boing = 0;
			break;
		}
	}
	
	return(boing);
	
}



